# -*- coding: utf-8 -*-
"""
Created on Wed Aug 26 20:07:43 2020

@author: 赵匡是
"""
from .exceptions import ModeError, LackOfParameterError, BrokenTimestepError,\
    KFoldError, LackOfLabelDictError, TaskError
from .labeldict import BaseDict

import math
import json
import random
import logging
import numpy as np
import tableprint as tp
from sklearn.model_selection import ShuffleSplit
from typing import Dict, Union

def flatten(seq):
    return [i for a in seq for i in a]

class Sample(object):
    def __init__(self, unit='epoch', tmin=0, tmax=0, n_splits=10,
                 test_size=0.1, class_balance=True, data_balance=False,
                 epoch_padding=False, data_padding=None,
                 task='classification', concat=False):
        self.__editable = True
        self.__autoencoder = False
        self.set_unit(unit)
        self.set_timespan(tmin, tmax)
        self.set_n_splits(n_splits)
        self.set_test_size(test_size)
        self.set_class_balance(class_balance)
        self.set_data_balance(data_balance)
        self.set_epoch_padding(epoch_padding)
        self.set_data_padding(data_padding)
        self.set_task(task)
        self.array_type = int
        self.concat = concat

    def set_unit(self, unit):
        '''
        Set the primary unit of sampling process. 
        
        The `Sample` object is to sample \'data\' and \'labels\' from given
        `Dataset` object,i.e. the `x`s and `y`s of a model. There are a lot
        of ways. 
        
        For example, using data of one epoch as `x` and the state of that
        epoch as `y`. Relatively, we don't care about which data that epoch
        come from. So the \'primary unit\' in this situation is \'epoch\'.
        
        In another example, we use a series of epochs as `x`, and just one
        label for the series. Epochs are not divisible in this situation. So
        the \'primary unit\' is \'data\'.
        
        If `unit == \'epoch\'` , the first dimension of sampled `x` will be
        \'epoch\'. Then if you use `condition` as sampled `y`, 
        '''
        self.__check_editable()
        self.unit = unit

    def get_unit(self):
        return self.unit

    def set_timespan(self, tmin, tmax):
        self.__check_editable()
        self.tmin = tmin
        self.tmax = tmax

    def get_timespan(self):
        return (self.get_tmin(), self.get_tmax())

    def get_tmax(self):
        return self.tmax

    def get_tmin(self):
        return self.tmin

    def set_n_splits(self, n_splits):
        '''
        Set the iterations of dataset.

        Parameters
        ----------
        n_splits : int
            Number of re-shuffling & splitting iterations.
            If you don\'t want to cross validation, you can set `n_splits=1`.

        '''
        self.__check_editable()
        self.n_splits = n_splits

    def get_n_splits(self):
        return self.n_splits

    def set_test_size(self, test_size):
        '''
        Set size(proportion) of the test set.

        Parameters
        ----------
        test_size : float
            Should be between 0.0 and 1.0 and represent the proportion
            of the dataset to include in the test split.

        '''
        self.__check_editable()
        self.test_size = test_size

    def get_test_size(self):
        return self.test_size

    def set_class_balance(self, class_balance):
        self.__check_editable()
        self.class_balance = class_balance

    def get_class_balance(self):
        return self.class_balance

    def set_data_balance(self, data_balance):
        self.__check_editable()
        if data_balance and self.unit == 'data':
            self.data_balance = False
            raise ValueError('you cannot set `data_balance` as True while '
                             'primary unit is `data`. Please use '
                             '`Sample.set_unit(\'epoch\')` first, then use '
                             '`Sample.set_data_balance(True)` again')
        else:
            self.data_balance = data_balance

    def get_data_balance(self):
        return self.data_balance

    def set_epoch_padding(self, epoch_padding):
        self.__check_editable()
        self.epoch_padding = epoch_padding

    def set_data_padding(self, data_padding, max_len = None):
        self.__check_editable()
        if data_padding is None:
            if self.unit == 'epoch':
                self.data_padding = False
            elif self.unit == 'data':
                self.data_padding = True
        else:
            self.data_padding = data_padding
        self.max_len = max_len

    def set_task(self, task):
        '''
        Set task you want to train. `'classification'`, `regression` and
        `autoencoder` is valid. Different task types use different sample
        mechanisms. For example, in `regression` task, we have no idea to do
        `class_balance`.

        Parameters
        ----------
        task : str
            'classification', 'regression' or 'autoencoder'.

        Raises
        ------
        ValueError
            Invalid parameter.

        '''
        candidate = ['classification', 'regression', 'autoencoder']
        # check tsk
        if task.lower() not in candidate:
            raise ValueError('invalid task.')
        if task == 'autoencoder':
            self.__autoencoder = True
            self.task = 'regression'
        else:
            self.task = task.lower()

    def set_label_array_type(self, array_type:type):
        '''
        Set the type of one-hot or N-hot label array the sample will return.
        If `array_type` is `None`, the function will return a number directly,
        i.e., not in terms of one-hot or N-hot array.

        Parameters
        ----------
        array_type : type or None, optional
            The type of elements in return array. The default and most
            recommended is int. You should input directly a `type`, not a
            `str`.
            For example:
                array_type=bool          (Correct)
                array_type='bool'        (Wrong)

            If `array_type` is `None`, the function will return a number
            directly, i.e., not in terms of one-hot or N-hot array.

        Raises
        ------
        TypeError
            Raised when you input an empty `label` parameter.

        Returns
        -------
        int or np.ndarray
            return `int` if array_type is None.
            else return one-hot or N-hot `np.ndarray`. 

        '''
        self.array_type = array_type

    def set_concat(self, concat:bool=True):
        '''
        Decide the output shape when tmin!=0 or tmax!=0, i.e., multiple time
        steps are sampled.
        If you have not set the `tmin` or `tmax`. You do not need to care about
        this parameter.
        The default is False, means the output will not concatenate time steps
        (i.e. sampled epochs) but leave the output shape as
                   `(time_step, channel[, feature-related])`
        In this situation, every
        time step occupies a unique matrix. It is suitable for RNN-related
        models.
        If `concat==True`, all the sampled epoches will concatenate by the
        'time' axis. It results in continuous feature which is suitable for TCN
        or other CNN-related models.

        Parameters
        ----------
        concat : bool, optional
            The default is True.

        '''
        self.concat = concat

    def set_x(self, x, labeldict: Dict[str, BaseDict] = {}):
        # 不强求类型。可以是dataset的label甚至是condition(比如用很多condition来判断疾病(age, sex)→diagnose
        self.__check_editable()
        if isinstance(x, str):
            self.x = [x]
        elif isinstance(x, list):
            self.x = x
        if self.__autoencoder:
            self.set_y(self.get_x())
        self.x_dict = labeldict

    def get_x(self):
        try:
            return self.x
        except:
            return None
            
    def set_y(self, y, labeldict: Dict[str, BaseDict] = {}):
        # unit是data就不是epoched，unit是epoch就是epoched。或者说，如果unit是data，就从condition里找，如果unit是epoch就从label里找。
        self.__check_editable()
        if isinstance(y, str):
            y = [y]
        elif isinstance(y, list):
            pass
        if self.__autoencoder and self.get_x() is not None and y != \
            self.get_x(): # assert x == y when using autoencoder
            raise ValueError('When using autoencoder, `y` must be `x` itself.')
        self.y = y
        self.y_dict = labeldict

    def get_y(self):
        try:
            return self.y
        except:
            return None

    def summary(self):
        # 打印整个sample的规则
        headers = ['Parameter', 'Value']
        content = [['unit', self.get_unit()],
                   ['tmin', self.get_tmin()],
                   ['tmax', self.get_tmax()],
                   ['n_splits', self.get_n_splits()],
                   ['test_size', self.get_test_size()],
                   ['class_balance', 'True' if self.get_class_balance() else
                    'False'],
                   ['data_balance', 'True' if self.get_data_balance() else
                    'False'],
                   ['epoch_padding', 'True' if self.epoch_padding else
                    'False'],
                   ['data_padding', 'True' if self.data_padding else 'False'],
                   ['max_len', self.max_len if self.max_len is not None else
                    'Not Set'],
                   ['task', self.task],
                   ['x', str(self.get_x())],
                   ['y', str(self.get_y())],
                   ['Sample Mode', self.mode if hasattr(self, 'mode') else 
                    'Not Set'],
                   ['label_array_type', '{0}: {1}'.format(
                       'one-hot/N-hot' if self.array_type else 'numeric',
                       self.array_type if self.array_type else 'int'
                   )]]
        tp.banner('Your Sampling Settings')
        tp.table(content, headers)

    @staticmethod
    def load(fpath):
        content = None
        with open(fpath, 'r') as f:
            content = json.load(f)
        sample = Sample(unit=content['unit'], tmin=content['tmin'],
                        tmax=content['tmax'], n_splits=content['n_splits'],
                        test_size=content['test_size'],
                        class_balance=content['class_balance'],
                        data_balance=content['data_balance'],
                        epoch_padding=content['epoch_padding'],
                        data_padding=content['data_padding'],
                        task=content['task'])
        sample.set_data_padding(content['data_padding'],
                                max_len= None if content['max_len'] == \
                                    'Not Set' else content['max_len'])
        sample.set_x(content['x'])
        sample.set_y(content['y'])
        return sample

    def save(self, fpath):
        if self.__autoencoder:
            task = 'autoencoder'
        else:
            task = self.task
        content = {'unit' : self.get_unit(),
                   'tmin' : self.get_tmin(),
                   'tmax' : self.get_tmax(),
                   'n_splits' : self.get_n_splits(),
                   'test_size' : self.get_test_size(),
                   'class_balance' : self.get_class_balance(),
                   'data_balance' : self.get_data_balance(),
                   'epoch_padding' : self.epoch_padding,
                   'data_padding' : self.data_padding,
                   'max_len' : self.max_len if self.max_len is not None else \
                    'Not Set',
                   'x' : self.get_x(),
                   'y' : self.get_y(),
                   'task' : task}
        with open(fpath, 'w') as f:
            json.dump(content, f)

    def __check_editable(self):
        if not self.__editable:
            raise AssertionError('You cannot edit parameters of this object '
                                 'after specifying dataset. In another words, '
                                 'any change after using `Sample.from_dataset`'
                                 ' is not allowed.')

    def __check_mode(self, current_mode):
        if current_mode != self.mode:
            if current_mode == 'train':
                raise ModeError('You cannot use `train_set()` or `test_set()` '
                                'in `predict` mode. Please try `Sample.sample`'
                                '.')
            elif current_mode == 'predict':
                raise ModeError('You cannot use `sample()` in `train` mode. '
                                'Please try `Sample.train_set()` and '
                                '`Sample.test_set()` to sample them '
                                'respectively.')

    def __check_preparation(self):
        # 检查subgroup和split有没有做
        if not self.__preparation:
            self.init()

    def __check_labeldict(self, dataset):
        if not self.__autoencoder:
            for x in self.get_x():
                if dataset.elements[x] == 'label' and x not in \
                    dataset.labeldict:
                    raise LackOfLabelDictError('The element `{0}` does not '
                                               'have a labeldict. Please use '
                                               '`Dataset.one_hot()` function '
                                               'to create one or manually set '
                                               'one.'.format(x))
            for y in self.get_y():
                if y not in dataset.labeldict:
                    raise LackOfLabelDictError('The element `{0}` does not '
                                               'have a labeldict. Please use '
                                               '`Dataset.one_hot()` function '
                                               'to create one or manually set '
                                               'one.'.format(y))

    def __check_iteration(self):
        '''
        In `train` mode, before use `Sample.train_set()` or `Sample.test_set()`
        , we should get one fold first (or a subset for cross validation).
        Unless the cross validation is not needed (when `n_splits == 1`).
        
        If `n_splits != 1`, you should use `Sample.next_fold()` every iteration
        to initialize train_set and test_set.

        Raises
        -------
        KFoldError
            Raised when `Sample.next_fold()` does not be used properly.

        '''
        # look up if all the thing needed exist
        # that is, `self.train` & `self.test`
        if (not hasattr(self, 'train') or not hasattr(self, 'test')) and \
            self.n_splits > 1:
            raise KFoldError('you\'ve set `n_splits = {0}`, so you have {0}'
                             ' slices of `train_set` and `test_set`. You must '
                             'set an iteration by using `Sample.next_fold()`'
                             ' to initialize the train_set and test_set for '
                             'this slice.'.format(self.get_n_splits()))
        if (not hasattr(self, 'train') or not hasattr(self, 'test')) and \
            self.n_splits == 1:
            self.next_fold()

    def __check_dataset(self, dataset):
        # check `x`
        if self.get_x() is None:
            raise LackOfParameterError('`x` hasn\'t set. Please use `set_x` '
                                       'first.')
        else:
            for item in self.get_x():
                # check if `x` in dataset.elements
                if item not in dataset.elements.keys():
                    raise AssertionError('The target dataset has no element `'+
                                         item + '`.')
                # while unit == `epoch`, `x` cannot be `condition` (because
                # `condition` is not epoched feature, cannot represent an epoch)
                elif self.get_unit() == 'epoch' and \
                    dataset.elements[item] == 'condition':
                    raise AssertionError('while unit == `epoch`, `x` '
                                         'cannot be `condition` (because '
                                         '`condition` is not epoched '
                                         'feature, cannot represent an '
                                         'epoch)')
        # check `y`
        if self.mode == 'train':  # in training mode, `y` must provided
            if self.get_y() is None:
                raise LackOfParameterError('`y` hasn\'t set. '
                                           'Please use `set_y` first.')
            else:
                for item in self.get_y():
                    # check if `y` in dataset.elements
                    if item not in dataset.elements.keys():
                        raise AssertionError('The target dataset has no'
                                             ' element `' + item + '`.')
                    # check if `y` is `label` or `condition`
                    elif dataset.elements[item] != 'label' and \
                        dataset.elements[item] != 'condition':
                        if len(self.get_x()) == 1 and self.get_x() == \
                            self.get_y():
                            logging.info('Use mode AUTOENCODER.')
                            print('[INFO] Use mode AUTOENCODER.')
                            self.set_task('autoencoder')
                        else:
                            raise AssertionError('The `y` must be `label` or'
                                                 ' `condition`. But ' + item +
                                                 ' was a `' +
                                                 dataset.elements[item] +'`.')
                    # while unit == `data`, `y` must be `condition` (because
                    # `label` is epoched label, cannot represent a data)
                    elif self.get_unit() == 'data' and \
                        dataset.elements[item] == 'label':
                            raise AssertionError('while unit == `data`, `y` '
                                                 'must be `condition` (because'
                                                 ' `label` is epoched label, '
                                                 'cannot represent a data)')
        # and in prediction mode, `y` will not be used

    def from_dataset(self, dataset, data_selection:dict=None,
                     mode:str='train'):
        '''
        Bind your sampling plan with a `Dataset`, which enables you begin to
        sample, but also disallows you to change your sampling plan again.

        Parameters
        ----------
        dataset : sleepology.dataset.Dataset
            A `Dataset` object. You should first create a `Dataset` from source
            or load a `Dataset` from disk, then input the `Dataset` as a
            parameter.

        data_selection : dict, optional
            The data or epochs you want to sample. The default is None which
            means use all the data or epochs.
            You should input a dict whose keys are `element_name` and values
            are `element_value`.
            For example, you have a `Dataset`, it will be sampled in unit of
            epochs. And it has `PSG` as an element of feature, and `LABEL` as
            an element of label, and also `DIAGNOSE` as an element of
            condition. You just want to sample epochs whose `LABEL` is `1` or 
            '2' and `DIAGNOSE` is `'healthy'`, you can use:
            >>> data_selection = {'LABEL' : [1, 2].
            ...                   'DIAGNOSE' : ['healthy']}

        mode : {'train', 'predict'}, optional
            Define the purpose why you sample. If you want to train a model,
            the sampled data will be balanced and split into train set and
            test set. Else you want to use an existed model to predict new
            data, the `Sample` process should not do any change to data, and it
            will definitely no `Sample.y()` because the correct answer has not
            provided. The default is 'train'.

        Raises
        ------
        ValueError

        Returns
        -------
        None.

        '''
        # set mode
        try:
            if mode.lower() == 'train':
                self.mode = 'train'
            elif mode.lower() == 'predict':
                self.mode = 'predict'
                self.y = None
                if self.data_padding and self.max_len is None:
                    raise ValueError('In mode `predict`, max_len of '
                                     'data_padding must be set.')
            else:
                raise ValueError('Invalid mode. Must be \'train\' or '
                                 '\'predict\'.')

            # check dataset
            self.__check_dataset(dataset)
            self.__check_labeldict(dataset)
        except Exception as e:
            del self.mode
            raise e

        # set dataset
        self.dataset = dataset
        self.dataset.shape_check()
        # === SELECT DATA ===
        # # use all the data or epochs
        if data_selection is None:
            if self.get_unit() == 'epoch':
                self.__selection = None
                self.data_selection = self.dataset.select_epochs()
            elif self.get_unit() == 'data':
                self.__selection = None
                self.data_selection = self.dataset.select_data()
        # # select data
        else:
            # data_selection has parameter in form of dict
            self.__selection = data_selection
            self.data_selection = set()
            # case: epoch
            if self.get_unit() == 'epoch':
                for e in data_selection.keys():  # for each element
                    # [1] there is already a subset of epochs in the
                    # `self.data_selection` and the element `e` provides some
                    # selection range.
                    if self.data_selection and len(self.__selection[e]) != 0:
                        self.data_selection.intersection_update(
                            set(self.dataset.select_epochs(e, data_selection[e]))
                        )
                    # [2] `self.data_selection` is an empty set
                    elif (not self.data_selection) and \
                        len(data_selection[e]) != 0:
                        self.data_selection = \
                            set(self.dataset.select_epochs(e, data_selection[e]))
                    # [3] other situation
                    elif self.data_selection:
                        pass
                    else:
                        self.data_selection = \
                            self.dataset.select_epochs()
            # case: data
            elif self.get_unit() == 'data':
                for e in data_selection.keys():  # for each element
                    if self.data_selection and len(data_selection[e]) != 0:
                        self.data_selection.intersection_update(
                            set(self.dataset.select_data(e, data_selection[e]))
                        )
                    elif (not self.data_selection) and \
                        len(data_selection[e]) != 0:
                        self.data_selection = \
                            self.dataset.select_data(e, data_selection[e])
                    elif self.data_selection:
                        pass
                    else:
                        self.data_selection = \
                            self.dataset.select_data()
        # set max_len
        if self.data_padding and self.max_len is None:
            epochs = self.dataset.epochs_per_data(self.data_list())
            self.max_len = max(epochs)

        # print the summary
        self.summary()

        # lock the config
        self.__editable = False
        # init preparation
        self.__preparation = False

    def data_list(self):
        '''
        Get data list from all the epochs or data.

        Returns
        -------
        list
            The data list.

        '''
        rst = set()
        for item in self.data_selection:
            if isinstance(item, str):
                rst.add(item)
            else:
                rst.add(item[0])
        return list(rst)

    def init(self):
        '''
        Initialize all the steps before sampling, including padding and
        spliting the train_set and test_set.
        '''
        logging.debug('== CLEANING VARIABLES ==')
        if hasattr(self, 'train'):
            del self.train
        if hasattr(self, 'test'):
            del self.test
        if hasattr(self, 'onesamp'):
            del self.onesamp
        if hasattr(self, 'opc'):  # one per class
            del self.opc

        logging.info('== INITIALIZING SAMPLING ==')
        if self.task == 'classification':
            self.subgroups()
        self.shuffle_split()
        logging.info('== INITIALIZING SAMPLING ==')

    def subgroups(self):
        # Only used in classification tasks
        '''
        Split data or epochs by classes into subgroups.
        Generate `Sample.classes`, which is a `dict` of different classes.
        '''
                # this method only supports `classification` task, not `regression`
        # task. Here we check it.
        if self.task != 'classification':
            raise TaskError('Only `classification` task has subgroups. Please '
                            'check the returned table when doing '
                            '`Sample.from_dataset`.')
        logging.info('== DISCRIMINATE DIFFERENT CLASSES ==')
        if self.__selection:
            element = self.__selection
        else:
            element = self.get_y()
        classes = self.dataset.stat_classes(element, self.get_unit())
        self.classes = {}
        for c in classes:
            self.classes[c] = [i for i in classes[c] 
                               if i in self.data_selection]  # intersection of
                                                             # data_selection
                                                             # and classes
        logging.info('== DISCRIMINATE DIFFERENT CLASSES ==')

    def shuffle_split(self):
        '''
        Generate generators for each of classes.

        After using `Sample.shuffle_split`, subgroups will be split into 
        `Sample.get_n_splits()` pieces, and in each piece 
        `Sample.get_test_size()` of the data will act as test set.

        '''
        logging.info('== SHUFFLE AND SPLIT DATASET ==')
        logging.info('n_splits = {0}'.format(self.get_n_splits()))
        logging.info('test_size = {0}'.format(self.get_test_size()))
        ss = ShuffleSplit(n_splits=self.get_n_splits(),
                          test_size=self.get_test_size())
        self.__current_fold = iter(range(1, self.get_n_splits() + 1))
        if self.task == 'classification':
            self.__k_fold = {}
            for c in self.classes:
                self.__k_fold[c] = ss.split(self.classes[c])
        else:
            self.__k_fold = ss.split(self.data_selection)
        logging.info('== SHUFFLE AND SPLIT DATASET ==')

    def next_fold(self, shuffle=True, disturb=False):
        self.__check_preparation()
        # start a new iteration
        current_fold = next(self.__current_fold)
        logging.info('== GET FOLD {0} =='.format(current_fold))
        train_set = []
        test_set = []
        self.disturb = disturb
        debug_classes = 0
        debug_train_set_len = 0
        debug_test_set_len = 0
        if self.task == 'regression':
            train, test = next(self.__k_fold)
            train = [self.data_selection[i] for i in train]
            test = [self.data_selection[i] for i in test]
            train_set.extend(train)
            test_set.extend(test)
        else:
            # append train_set and test_set of each class
            for c in self.__k_fold:
                debug_classes += 1
                train, test = next(self.__k_fold[c])
                train = [self.classes[c][i] for i in train]
                test = [self.classes[c][i] for i in test]
                train_set.append(train)
                test_set.append(test)
                debug_train_set_len += len(train)
                debug_test_set_len += len(test)
            ## train_set is a 2-D list. The first dimension is `class`. The
            ## second dimension is `data` or `(data, epoch)` tuple.
            logging.info('Dataset has ' + str(debug_classes) + ' classes')
            logging.info('Current train_set length ' + str(debug_train_set_len))
            logging.info('Current test_set length ' + str(debug_test_set_len))
            if self.get_class_balance() and not self.get_data_balance():
                logging.info('Balancing classes ...')
                # find the max_len of different classes. Then oversample other
                # class to the max_len
                train_len = max(len(i) for i in train_set)
                logging.info('The max_len of train_set is ' + str(train_len))
                test_len = max(len(i) for i in test_set)
                logging.info('The max_len of test_set is ' + str(test_len))
                # balance every class
                for pcs in range(len(train_set)):  # each piece is a class
                    idx = np.random.choice(range(len(train_set[pcs])), train_len)
                    train_set[pcs] = [train_set[pcs][i] for i in idx]
                    logging.debug('[{0}] sample train_set to {1}.'.format(str(pcs),
                        str(len(train_set[pcs]))))
                for pcs in range(len(test_set)):  # each piece is a class
                    idx = np.random.choice(range(len(test_set[pcs])), test_len)
                    test_set[pcs] = [test_set[pcs][i] for i in idx]
                    logging.debug('[{0}] sample test_set to {1}.'.format(str(pcs),
                        str(len(test_set[pcs]))))
                train_set = flatten(train_set)
                test_set = flatten(test_set)
            elif self.get_data_balance() and not self.get_class_balance():
                logging.info('Balancing data ...')
                # this condition assert self.unit == 'epoch'
                # don't care about classes. Concatenate classes
                train_set = flatten(train_set)
                logging.info('Flatten train_set, length: ' + len(train_set))
                test_set = flatten(test_set)
                logging.info('Flatten test_set, length: ' + len(test_set))
                # and discriminate train_set and test_set by `data_name`
                train_set = list(self.__discriminate_data(train_set).values())
                test_set = list(self.__discriminate_data(test_set).values())
                ## Now, the train_set is also a 2-D list. The first dimension is
                ## `data`. The second dimension is `(data, epoch)` tuple.
                # find the max_len of different data. Then oversample other data
                # to the max_len
                train_len = max(len(i) for i in train_set)
                logging.info('The max_len of train_set is ' + str(train_len))
                test_len = max(len(i) for i in test_set)
                logging.info('The max_len of test_set is ' + str(test_len))
                # balance every data
                for pcs in range(len(train_set)):  # each piece is a datum
                    idx = np.random.choice(range(len(train_set[pcs])), train_len)
                    train_set[pcs] = [train_set[pcs][i] for i in idx]
                    logging.debug('[{0}] sample train_set to {1}.'.format(str(pcs),
                        str(len(train_set[pcs]))))
                for pcs in range(len(test_set)):  # each piece is a datum
                    idx = np.random.choice(range(len(test_set[pcs])), test_len)
                    test_set[pcs] = [test_set[pcs][i] for i in idx]
                    logging.debug('[{0}] sample test_set to {1}.'.format(str(pcs),
                        str(len(test_set[pcs]))))
                train_set = flatten(train_set)
                test_set = flatten(test_set)
            elif self.get_class_balance() and self.get_data_balance():
                logging.info('Balancing data and balancing classes ...')
                logging.info('> Discriminate different data in classes. So it '
                             'generate a (class, data) structure')
                tem = []
                for ts in train_set:
                    ## extending a 2-D list. The first dimension is `data`. The
                    ## second dimension is `(data, epoch)` tuple.
                    tem.extend(list(self.__discriminate_data(ts).values()))
                train_set = tem
                tem = []
                for tt in test_set:
                    ## extending a 2-D list. The first dimension is `data`. The
                    ## second dimension is `(data, epoch)` tuple.
                    tem.extend(list(self.__discriminate_data(tt).values()))
                test_set = tem
                ## the train_set and test_set remain a 2-D array because we have
                ## used `extend` instead of `append`. The first dimension is a
                ## combination of `data` and `class`. A list of certain
                ## `(data, class)` in the second dimension.
                # compute max_len of train_set and test_set
                train_len = max(len(i) for i in train_set)
                logging.info('The max_len of train_set is ' + str(train_len))
                test_len = max(len(i) for i in test_set)
                logging.info('The max_len of test_set is ' + str(test_len))
                # balance every (data, class)
                for pcs in range(len(train_set)):
                    idx = np.random.choice(range(len(train_set[pcs])), train_len)
                    train_set[pcs] = [train_set[pcs][i] for i in idx]
                    logging.debug('[{0}] sample train_set to {1}.'.format(str(pcs),
                        str(len(train_set[pcs]))))
                for pcs in range(len(test_set)):
                    idx = np.random.choice(range(len(test_set[pcs])), test_len)
                    test_set[pcs] = [test_set[pcs][i] for i in idx]
                    logging.debug('[{0}] sample test_set to {1}.'.format(str(pcs),
                        str(len(test_set[pcs]))))
                train_set = flatten(train_set)
                test_set = flatten(test_set)
            else:
                train_set = flatten(train_set)
                test_set = flatten(test_set)
        # finally
        if shuffle:
            logging.info('shuffle...')
            random.shuffle(train_set)
            random.shuffle(test_set)
        self.train = train_set
        self.test = test_set
        # disturb, which disturb the match of x and y, i.e. shuffle the y
        if disturb:
            train_set = train_set.copy()
            test_set = test_set.copy()
            random.shuffle(train_set)
            random.shuffle(test_set)
            self.train_y = train_set
            self.test_y = test_set
        logging.info('Length of `train_set` this fold: {0}'.format(
            len(self.train)))
        logging.info('Length of `test_set` this fold: {0}'.format(
            len(self.test)))
        logging.info('== GET FOLD {0} =='.format(current_fold))
        return current_fold

    def __discriminate_data(self, ori):
        '''
        Discriminate different data in a series of epochs.
        
        The input is a list whose element has a form `(data_name, epoch)`.
        The output is dict:
            output = {data_name: [(data_name, epoch), (data_name, epoch), ...],
                      data_name: [(data_name, epoch), (data_name, epoch), ...],
                      ...}
        Example:
        >>> list(Sample.__discriminate_data(ori).keys())  # acquire a data list
        
        >>> list(Sample.__discriminate_data(ori).values())  # split different
                                                            # data

        Parameters
        ----------
        ori : list
            A list whose element has a form `(data_name, epoch)`.

        Returns
        -------
        rst : dict
            A dict, in form of:
            output = {data_name: [(data_name, epoch), (data_name, epoch), ...],
                      data_name: [(data_name, epoch), (data_name, epoch), ...],
                      ...}

        '''
        rst = {}
        for t in ori:
            if t[0] not in rst:
                rst[t[0]] = []
            rst[t[0]].append(t)
        return rst

    def train_set(self, generator=True):
        if generator:
            logging.info('Use generator')
            return self.train_set_generator()
        else:
            logging.info('Do not use generator')
            x_samp = []
            y_samp = []
            for x, y in self.train_set_generator():
                x_samp.append(x)
                y_samp.append(y)
            x_samp = np.asarray(x_samp)
            y_samp = np.asarray(y_samp)
            return (x_samp, y_samp)

    def train_set_generator(self):
        self.__check_mode('train')
        self.__check_iteration()
        for idx, item in enumerate(self.train):
            try:
                if self.get_unit() == 'epoch' and not self.disturb:
                    yield self.dataset.sample_epoch(
                        item[0],
                        item[1],
                        (self.get_x(), self.get_y()),
                        tmin=self.get_tmin(),
                        tmax=self.get_tmax(), 
                        epoch_padding=self.epoch_padding,
                        autoencoder=self.__autoencoder,
                        array_type=self.array_type,
                        concat=self.concat,
                        x_dict = self.x_dict,
                        y_dict = self.y_dict)
                elif self.get_unit() == 'data' and not self.disturb:
                    yield self.dataset.sample_data(
                        item,
                        (self.get_x(), self.get_y()),
                        tmin=self.get_tmin(),
                        tmax=self.get_tmax(),
                        data_padding=self.data_padding,
                        max_len=self.max_len,
                        epoch_padding=self.epoch_padding,
                        autoencoder=self.__autoencoder,
                        array_type=self.array_type,
                        concat=self.concat,
                        x_dict = self.x_dict,
                        y_dict = self.y_dict)
                elif self.get_unit() == 'epoch' and self.disturb:
                    yield self.dataset.sample_epoch(
                        item[0],
                        item[1],
                        (self.get_x(), self.get_y()),
                        tmin=self.get_tmin(),
                        tmax=self.get_tmax(), 
                        epoch_padding=self.epoch_padding,
                        test_data_name=self.train_y[idx][0],
                        test_epoch=self.train_y[idx][1],
                        autoencoder=self.__autoencoder,
                        array_type=self.array_type,
                        concat=self.concat,
                        x_dict = self.x_dict,
                        y_dict = self.y_dict)
                elif self.get_unit() == 'data' and not self.disturb:
                    yield self.dataset.sample_data(
                        item,
                        (self.get_x(), self.get_y()),
                        tmin=self.get_tmin(),
                        tmax=self.get_tmax(),
                        data_padding=self.data_padding,
                        max_len=self.max_len,
                        epoch_padding=self.epoch_padding,
                        test_data_name=self.train_y[idx],
                        autoencoder=self.__autoencoder,
                        array_type=self.array_type,
                        concat=self.concat,
                        x_dict = self.x_dict,
                        y_dict = self.y_dict)
            except BrokenTimestepError:
                continue

    def test_set(self, generator=True):
        if generator:
            logging.info('Use generator')
            return self.test_set_generator()
        else:
            logging.info('Do not use generator')
            x_samp = []
            y_samp = []
            for x, y in self.test_set_generator():
                x_samp.append(x)
                y_samp.append(y)
            x_samp = np.asarray(x_samp)
            y_samp = np.asarray(y_samp)
            return (x_samp, y_samp)

    def test_set_generator(self):
        self.__check_mode('train')
        self.__check_iteration()
        for idx, item in enumerate(self.test):
            try:
                if self.get_unit() == 'epoch' and not self.disturb:
                    yield self.dataset.sample_epoch(
                        item[0],
                        item[1],
                        (self.get_x(), self.get_y()),
                        tmin=self.get_tmin(),
                        tmax=self.get_tmax(), 
                        epoch_padding=self.epoch_padding,
                        autoencoder=self.__autoencoder,
                        array_type=self.array_type,
                        concat=self.concat,
                        x_dict = self.x_dict,
                        y_dict = self.y_dict)
                elif self.get_unit() == 'data' and not self.disturb:
                    yield self.dataset.sample_data(
                        item,
                        (self.get_x(), self.get_y()),
                        tmin=self.get_tmin(),
                        tmax=self.get_tmax(),
                        data_padding=self.data_padding,
                        max_len=self.max_len,
                        epoch_padding=self.epoch_padding,
                        autoencoder=self.__autoencoder,
                        array_type=self.array_type,
                        concat=self.concat,
                        x_dict = self.x_dict,
                        y_dict = self.y_dict)
                elif self.get_unit() == 'epoch' and self.disturb:
                    yield self.dataset.sample_epoch(
                        item[0],
                        item[1],
                        (self.get_x(), self.get_y()),
                        tmin=self.get_tmin(),
                        tmax=self.get_tmax(), 
                        epoch_padding=self.epoch_padding,
                        test_data_name=self.test_y[idx][0],
                        test_epoch=self.test_y[idx][1],
                        autoencoder=self.__autoencoder,
                        array_type=self.array_type,
                        concat=self.concat,
                        x_dict = self.x_dict,
                        y_dict = self.y_dict)
                elif self.get_unit() == 'data' and self.disturb:
                    yield self.dataset.sample_data(
                        item,
                        (self.get_x(), self.get_y()),
                        tmin=self.get_tmin(),
                        tmax=self.get_tmax(),
                        data_padding=self.data_padding,
                        max_len=self.max_len,
                        epoch_padding=self.epoch_padding,
                        test_data_name=self.test_y[idx],
                        autoencoder=self.__autoencoder,
                        array_type=self.array_type,
                        concat=self.concat,
                        x_dict = self.x_dict,
                        y_dict = self.y_dict)
            except BrokenTimestepError:
                continue

    def one(self, generator:bool=True):
        '''
        Sample only one data and its label in dataset. This one data will
        be repeatedly sampled and fed into the model. This method is designed
        for testing a model. If the model cannot fit just one sample, there
        maybe some problems with the model, such as gradient disappearance.

        Parameters
        ----------
        generator : bool, optional
            Determine the return type. If True, the returned object is a
            generator, it is supported by tensorflow and memory-saving. If
            False, the returned object is a np.ndarray, it is supported by
            scikit-learn.
            The default is True.

        Returns
        -------
        generator or np.ndarray

        '''
        if generator:
            logging.info('Use generator')
            return self.one_generator()
        else:
            logging.info('Do not use generator')
            x_samp = []
            y_samp = []
            for x, y in self.one_generator():
                x_samp.append(x)
                y_samp.append(y)
            x_samp = np.asarray(x_samp)
            y_samp = np.asarray(y_samp)
            return (x_samp, y_samp)

    def one_generator(self):
        '''
        Sample only one data and its label in dataset. This one data will
        be repeatedly sampled and fed into the model. This method is designed
        for testing a model. If the model cannot fit just one sample, there
        maybe some problems with the model, such as gradient disappearance.

        Yields
        ------
        np.ndarray

        '''
        # the purpose of `Sample.one` is only to get one sample, regardless of
        # whether the mode `train` or `test`, or whether the dataset needs to
        # cross validation or not.
        if not hasattr(self, 'onesamp'):
            onesamp = random.sample(self.data_selection, 1)
            self.onesamp = onesamp * len(self.data_selection)
        for idx, item in enumerate(self.onesamp):
            try:
                if self.get_unit() == 'epoch':
                    yield self.dataset.sample_epoch(
                        item[0],
                        item[1],
                        (self.get_x(), self.get_y()),
                        tmin=self.get_tmin(),
                        tmax=self.get_tmax(), 
                        epoch_padding=self.epoch_padding,
                        autoencoder=self.__autoencoder,
                        array_type=self.array_type,
                        concat=self.concat,
                        x_dict = self.x_dict,
                        y_dict = self.y_dict)
                elif self.get_unit() == 'data':
                    yield self.dataset.sample_data(
                        item,
                        (self.get_x(), self.get_y()),
                        tmin=self.get_tmin(),
                        tmax=self.get_tmax(),
                        data_padding=self.data_padding,
                        max_len=self.max_len,
                        epoch_padding=self.epoch_padding,
                        autoencoder=self.__autoencoder,
                        array_type=self.array_type,
                        concat=self.concat,
                        x_dict = self.x_dict,
                        y_dict = self.y_dict)
            except BrokenTimestepError:
                continue
    
    def one_per_class(self, generator=True):
        '''
        Sample only one data and its label per class. These data will be
        repeatedly sampled and fed into the model. This method is designed
        for testing a model. If the model cannot fit such small dataset, there
        maybe some problems with the model, such as gradient disappearance.

        Parameters
        ----------
        generator : bool, optional
            Determine the return type. If True, the returned object is a
            generator, it is supported by tensorflow and memory-saving. If
            False, the returned object is a np.ndarray, it is supported by
            scikit-learn.
            The default is True.

        Returns
        -------
        generator or np.ndarray

        '''
        # this method only supports `classification` task, not `regression`
        # task. Here we check it.
        if self.task != 'classification':
            raise TaskError('Only `classification` task has classes. Please '
                            'check the returned table when doing '
                            '`Sample.from_dataset` if your task is '
                            'classification, or use `Sample.one()` instead '
                            'of `Sample.one_per_class()` in regression task,')
        if generator:
            logging.info('Use generator')
            return self.one_per_class_generator()
        else:
            logging.info('Do not use generator')
            x_samp = []
            y_samp = []
            for x, y in self.one_per_class_generator():
                x_samp.append(x)
                y_samp.append(y)
            x_samp = np.asarray(x_samp)
            y_samp = np.asarray(y_samp)
            return (x_samp, y_samp)

    def one_per_class_generator(self):
        '''
        Sample only one data and its label per class. These data will be
        repeatedly sampled and fed into the model. This method is designed
        for testing a model. If the model cannot fit such small dataset, there
        maybe some problems with the model, such as gradient disappearance.

        Yields
        ------
        np.ndarray

        '''
        if not hasattr(self, 'opc'):
            if not hasattr(self, 'classes'):
                self.subgroups()
            # the purpose of `Sample.one_per_class` is only to get samples,
            # regardless of whether the mode `train` or `test`, or whether the
            # dataset needs to cross validation or not.
            opc = [] # one_per_class
            for k in self.classes.keys():
                opc.extend(random.sample(self.classes[k], 1))
            # the length of samples = data_length
            self.opc = opc * math.floor(
                len(self.data_selection) / len(self.classes.keys()))
        for idx, item in enumerate(opc):
            try:
                if self.get_unit() == 'epoch':
                    yield self.dataset.sample_epoch(
                        item[0],
                        item[1],
                        (self.get_x(), self.get_y()),
                        tmin=self.get_tmin(),
                        tmax=self.get_tmax(), 
                        epoch_padding=self.epoch_padding,
                        autoencoder=self.__autoencoder,
                        array_type=self.array_type,
                        concat=self.concat,
                        x_dict = self.x_dict,
                        y_dict = self.y_dict)
                elif self.get_unit() == 'data':
                    yield self.dataset.sample_data(
                        item,
                        (self.get_x(), self.get_y()),
                        tmin=self.get_tmin(),
                        tmax=self.get_tmax(),
                        data_padding=self.data_padding,
                        max_len=self.max_len,
                        epoch_padding=self.epoch_padding,
                        autoencoder=self.__autoencoder,
                        array_type=self.array_type,
                        concat=self.concat,
                        x_dict = self.x_dict,
                        y_dict = self.y_dict)
            except BrokenTimestepError:
                continue

    def sample(self):
        self.__check_mode('predict')
        pass
    


# -*- coding: utf-8 -*-
import os
import json

class Procedure(object):
    '''
        用于生成并保存对于每一个通道的处理方式、以及保存哪些通道。（只保留通道而不经任何处理的，
        preprocessing_method为accept） 2020-2-18
        self.__proce = ['preprocessing_method', [channel_list], [param_list]]
        self.__proce[i]是每一个条目
        self.__proce[i][0]是方法名
        self.__proce[i][1]是通道列表
        self.__proce[i][2]是参数列表
    '''
    def __init__(self, args = None):
        '''
        2020-3-3
        args 可能有四种形式：
        ①list的话，可能是直接输入self.__proce列表
        ②function的话，可能是直接输入self.procedure函数(一定要输入高阶函数！)
        ③str的话，可能是存储Procedure的文件
        ④None的话，就是空白Procedure
        '''
        self.__proce = list()
        self.channel_list = set() # 2020-2-19 整理一个不重复的通道列表
        if args is not None:
            if isinstance(args, str):
                self.load(os.path.abspath(args))
            elif isinstance(args, list):
                self.set_procedure(args)
            elif callable(args):
                self.procedure = args
            else:
                raise TypeError('Procedure class connot be initialized by given args')
    
    def add_method(self, method, chan_list, param_list = None):
        # 2020-2-13 为已经存在的通道添加操作列表
        if not isinstance(method, str) or not isinstance(chan_list, list) \
            or not isinstance(param_list, list) and param_list is not None:
            raise TypeError()
        self.__proce.append([method, chan_list, param_list])
        for channel in chan_list:
            self.channel_list.add(channel)
    
    def clear(self):
        # 2020-2-19 重新初始化
        self.__proce = list()
        self.channel_list = set()
    
    def add_channels(self, channels):
        if not isinstance(channels, list):
            channels = [channels]
        for channel in channels:
            self.channel_list.add(channel)
    
    def get_channels(self):
        # 2020-2-13 获取已经存在的通道列表
        return list(self.channel_list)
    
    def set_procedure(self, proce): # 2020-2-19 直接接收格式化后的list，而不一个一个method添加
        self.__proce = proce
        # 添加通道列表 2020-2-19
        for method in self.__proce:
            for channel in method[1]:
                self.channel_list.add(channel)
    
    def get_procedure(self):
        return self.__proce
    
    def procedure(self):
        # 2020-3-1 高阶函数，返回一个处理流程。低阶函数的输入是raw，输出是np.ndarray
        ####################  未完待续  ####################
        def pipeline(data):
            for preprocessing_method, channel_list, param_list in self.__proce:
                pass
                #data = getattr(preprocess, preprocessing_method)(data, 
                #              param_list, picks = channel_list)
            return data
        return pipeline  
    
    def dumps(self):
        return json.dumps(self.__proce)
    
    def load(self, load_path):
        # 2020-2-13 从指定路径加载步骤
        with open(load_path, 'r') as f:
            self.__proce = json.load(fp = f)
            # 添加通道列表 2020-2-19
            for method in self.__proce:
                for channel in method[1]:
                    self.channel_list.add(channel)
    
    def save(self, save_path):
        # 2020-2-13 将步骤保存至指定路径
        with open(save_path, 'w') as f:
            json.dump(self.__proce, f)


# -*- coding: utf-8 -*-


class ModeError(Exception):
    '''
    When training, both `train_set` and `test_set` are required. And these
    dataset may be elaborately balanced and randomly shuffled. So please use
    `Sample.train_set()` and `Sample.test_set()`.

    However, when predicting, there are no `train_set` or `test_set`. And it is
    impossible to balance data or shuffle data. Please use `Sample.sample()` to
    just sample data without any other process.
    '''
    pass


class LackOfParameterError(Exception):
    '''
    Raised when sampling data. Different sample `unit` and sample `mode` needs
    different parameters. Please make sure your input meets the need of 
    `Sample` object.
    '''
    pass


class LackOfLabelDictError(Exception):
    '''
    Raised when you input dataset to a `Sample` instance and `y` element in
    that dataset does not have a label dict. Please use `Dataset.one_hot()`
    function to create one or manually set one.
    '''
    pass


class DataStateError(Exception):
    '''
    Raised when data in the dataset is not in a correct state, such as `ERROR`
    state. Please make sure you have tried effective `check` process and only
    input proper data into pipeline.
    '''
    pass


class BrokenTimestepError(Exception):
    '''
    Raised when sampling data in epochs. When epoch at edge cannot fill the
    required timestep to the full, and `epoch_padding == False`, the sample
    function will raise `BrokenTimestepError`, and the best way to handle this
    exception is using `continue` statement and going into next iteration to
    skip (also dispose) this epoch.
    '''
    pass


class KFoldError(Exception):
    '''
    Raised when iteration is not correctly set.
    '''
    pass


class TaskError(Exception):
    '''
    Raised when operation ordered in an improper task.
    '''
    pass

class LabelValueConflictError(Exception):
    '''
    Raised when a value represent more than one labels.
    '''
    pass


# -*- coding: utf-8 -*-
"""
Created on Sat Mar  7 11:31:23 2020

@author: Zhao Kuangshi
"""
from .source import *
from .utils import total_size
# from .procedure import Procedure
from .labeldict import BaseDict, AASM, ClassDict
from .exception_logger import ExceptionLogger
from .exceptions import DataStateError, BrokenTimestepError

import h5py
import tempfile
import numpy as np
import matplotlib.pyplot as plt
import math
import os
import logging
import traceback
from typing import Union, Dict


os.environ["HDF5_USE_FILE_LOCKING"] = 'FALSE'

package_root = os.path.dirname(os.path.abspath(__file__))

class Dataset(object):
    '''
    2020-2-20
    引入数据、加载数据、预处理数据、记录日志、*合并标签、*输出文件、*输出数组。
    '''
    # the state code
    ADDED = 0
    CHECKED = 1
    PREPROCESSED = 2
    PREDICTED = 3
    ERROR = -1
    
    def __init__(self, dataset_name, save_path, comment = '', 
                labeldict: Union[BaseDict, Dict[str, BaseDict], None] = None,
                mode = 'memory'):
        '''
        Create a new `Dataset()`

        Parameters
        ----------
        dataset_name : str
            The name of the dataset. It will be the filename if you save this
            dataset.
        save_path : path-like
            Where you want to save this dataset.
        comment : str, optional
            A comment string to describe your dataset. The default is ''.
        labeldict : instance of LabelDict OR a dictionary of LabelDict whose
            keys are label names, optional

            A dictionary to manage labels of dataset. It will translate human-
            readable label to the style which fits machine learning.
        mode : {'memory', 'disk'}, optional
            For the faster speed, 'memory' mode will process all the data in
            the memory. But if you have a huge dataset which exceeds or will
            exhaust your memory, the 'disk' mode will cache all the data at
            disk until you are to use them. It do slow down the computation,
            but it is a good resolution when you are not able to extend your
            memory. The default is 'memory'.

        '''
        self.VERSION = 0.2
        self.path = save_path
        if self.path is not None and (dataset_name is not None):
            self.set_name(dataset_name)
        self.comment = comment
        # set LabelDict
        self.labeldict = {}
        if isinstance(labeldict, BaseDict):
            self.set_labeldict('LABEL', labeldict)
        elif isinstance(labeldict, dict):
            for label in labeldict.keys():
                self.set_labeldict(label, labeldict[label])
        # set mode
        if mode.lower() == 'memory':
            self.disk_mode = False
        else:
            self.disk_mode = True
        self.__init_dataset()
        self.exception_logger = ExceptionLogger(self)
        self.elements = dict()
        self.features = set()
        self.labels = set()
        self.has_label = False

    def disk_file(self):
        if not hasattr(self, 'df'): # The cache not being created
            cache_dir = os.path.expanduser(os.path.join('~', '.sleepology'))
            if not os.path.exists(cache_dir):
                os.mkdir(cache_dir)
            tf = tempfile.TemporaryFile(dir = cache_dir)
            self.df = h5py.File(tf, 'w')
            return self.df
        else:
            return self.df

    def set_labeldict(self, label_name, labeldict):
        self.labeldict[label_name] = labeldict

    ### NAME ###

    def set_name(self, name):
        '''
        Set the name of the dataset.

        Parameters
        ----------
        name : str
            The name of the dataset.

        '''
        # 2020-2-23 改名字同时改保存路径
        self.name = name
        self.save_direction = os.path.join(os.path.abspath(self.path), 
                                           self.name + '.dataset')

    def get_name(self):
        '''
        Get the name of the dataset.

        Returns
        -------
        str
            The name of the dataset.

        '''
        return self.name

    ### COMMENT ###

    def set_comment(self, comment):
        '''
        Set the comment of the dataset. The comment is used to discriminate 
        datasets.

        Parameters
        ----------
        comment : str
            The comment.

        '''
        self.comment = comment

    def get_comment(self):
        '''
        Get the comment of the dataset. The comment is used to discriminate 
        datasets.

        Returns
        -------
        str
            The comment.

        '''
        return self.comment

    ### DATASET ###

    def __init_dataset(self):
        if self.disk_mode:
            self.__df_init_dataset(self.disk_file())
        else:
            self.__mm_init_dataset()

    def __mm_init_dataset(self):
        self.dataset = {}

    def __df_init_dataset(self, disk_file):
        disk_file.create_group('dataset')

    ### ELEMENT ###

    def __new_feature(self, feature_name):
        if feature_name in self.elements and self.elements[feature_name] != 'feature':
            raise ValueError('Feature name conflict. A `' +
                             self.elements[feature_name] + '` has same name.')
        elif feature_name in self.elements and self.elements[feature_name] == 'feature':
            pass
        else:
            self.features.add(feature_name)
            self.elements[feature_name] = 'feature'

    def __new_label(self, label_name):
        if label_name in self.elements and self.elements[label_name] != 'label':
            raise ValueError('Label name conflict. A `' +
                             self.elements[label_name] + '` has same name.')
        elif label_name in self.elements and self.elements[label_name] == 'label':
            pass
        else:
            self.labels.add(label_name)
            self.elements[label_name] = 'label'

    def __new_condition(self, condition_name):
        if condition_name in self.elements and self.elements[condition_name] != 'condition':
            raise ValueError('Condition name conflict. A `' +
                             self.elements[condition_name] + '` has same name.')
        if condition_name in self.elements and self.elements[condition_name] == 'condition':
            pass
        else:
            self.elements[condition_name] = 'condition'

    ### DATA ###

    def __init_new_data(self, data_name):
        if self.disk_mode:
            self.__df_init_new_data(self.disk_file(), data_name)
        else:
            self.__mm_init_new_data(data_name)

    def __mm_init_new_data(self, data_name):
        # initialize the new data
        self.dataset[data_name] = dict()
        self.dataset[data_name]['source'] = dict()
        self.dataset[data_name]['data'] = dict()

    def __df_init_new_data(self, disk_file, data_name):
        disk_file['dataset'].create_group(data_name)
        disk_file['dataset'][data_name].create_group('data')

    def add_data(self, feature_source, label_source = None, data_name = None):
        # format the source input
        if isinstance(feature_source, Source):
            feature_source = {'FEATURE': feature_source}
        if isinstance(label_source, Source):
            label_source = {'LABEL': label_source}
        # ensure the `data_name`. If the same name has existed, append
        # number behind
        main_source = list(feature_source.keys())[0]
        if data_name is None:
            data_name = feature_source[main_source].name
        if data_name in self.get_data() and (self.get_source(data_name,\
            main_source).path == feature_source[main_source].path): 
            pass
        else: 
            if data_name in self.get_data(): 
                i = 1
                while (data_name + '_' + str(i)) in self.get_data():
                    i += 1
                data_name = (data_name + '_' + str(i))

            self.__init_new_data(data_name)
            
            # register the source
            for feature_name in feature_source.keys():
                self.__new_feature(feature_name)
                self.__set_source(data_name, feature_name, 
                                  feature_source[feature_name]) 
            if label_source is not None:
                self.has_label = True
                for label_name in label_source.keys():
                    self.__new_label(label_name)
                    self.__set_source(data_name, label_name,
                                      label_source[label_name])
            
            self.__set_state(data_name, Dataset.ADDED)

    def del_data(self, data_name = None):
        '''
        Delete data from the dataset.

        Parameters
        ----------
        data_name : str, optional
            The data to be deleted. If None, all the data will be deleted.

        '''
        if self.disk_mode:
            self.__df_del_data(self.disk_file(), data_name)
        else:
            self.__mm_del_data(data_name)

    def __mm_del_data(self, data_name = None):
        '''
        Delete data from the dataset.

        Parameters
        ----------
        data_name : str, optional
            The data to be deleted. If None, all the data will be deleted.

        '''
        if data_name is None:
            self.dataset = {}
        else:
            self.dataset.pop(data_name)

    def __df_del_data(self, disk_file, data_name = None):
        if data_name is None:
            disk_file['dataset'].clear()
        else:
            disk_file['dataset'].pop(data_name)

    def get_data(self, data_name = None, attr = None):
        if self.disk_mode:
            return self.__df_get_data(self.disk_file(), data_name, attr)
        else:
            return self.__mm_get_data(data_name, attr)

    def __mm_get_data(self, data_name = None, attr = None):
        # 2020-3-1
        if data_name is None:
            return self.dataset.keys()
        elif attr is None:
            return self.dataset.get(data_name)
        else:
            return self.dataset.get(data_name).get(attr)

    def __df_get_data(self, disk_file, data_name = None, attr = None):
        if data_name is None:
            return disk_file['dataset'].keys()
        elif attr is None:
            return disk_file['dataset'][data_name]
        elif attr in disk_file['dataset'][data_name].attrs:
            return disk_file['dataset'][data_name].attrs[attr]
        else:
            return disk_file['dataset'][data_name][attr]

    def select_data(self, condition_type = None, condition_value = None):
        # 删除所有文件中指定label的数据
        if condition_type is None and condition_value is None:
            rst = set()
            for data_name in self.get_data():
                if self.__get_state(data_name) < Dataset.PREPROCESSED:
                    continue
                logging.info(data_name)
                rst.add(data_name)
        elif condition_type is not None and condition_value is not None:
            rst = set()
            if not isinstance(condition_value, list):
                condition_value = [condition_value]
            for l in condition_value:
                for data_name in self.get_data():
                    if self.__get_state(data_name) < Dataset.PREPROCESSED:
                        continue
                    logging.info(data_name)
                    if self.get_condition(data_name, condition_type) == l:
                        rst.add(data_name)
        else:
            raise TypeError('You should input all of or none of\n' + 
                            '`condition_type` and `condition_value`')
        return rst

    def exclude_data(self, condition_type, condition_value):
        if not isinstance(condition_value, list):
            condition_value = [condition_value]
        for l in condition_value:
            to_be_del = []
            for data_name in self.get_data():
                if self.__get_state(data_name) != Dataset.ERROR:
                    logging.info(data_name)
                    if self.get_condition(data_name, condition_type) == l:
                        to_be_del.append(data_name)
            for d in to_be_del:
                self.del_data(d)

    def epochs_per_data(self, data_name = None):
        r = []
        if data_name is None:
            data = self.get_data()
        elif isinstance(data_name, str):
            data = [data_name]
        elif isinstance(data_name, list):
            data = data_name
        for d in data:
            if self.__get_state(d) == Dataset.PREPROCESSED:
                r.append(len(self.get_epochs(d)))
        return r

    ### CONDITION ###
    def has_condition(self, data_name, condition_type = None):
        if self.disk_mode:
            return self.__df_has_condition(self.disk_file(),
                                           data_name, condition_type)
        else:
            return self.__mm_has_condition(data_name, condition_type)

    def __mm_has_condition(self, data_name, condition_type = None):
        if condition_type is None:
            return 'CONDITION' in self.dataset[data_name]
        elif 'CONDITION' in self.dataset[data_name]:
            return condition_type in self.dataset[data_name]['CONDITION']
        else:
            return False

    def __df_has_condition(self, disk_file, data_name, condition_type = None):
        if condition_type is None:
            return 'CONDITION' in disk_file['dataset'][data_name].keys()
        elif 'CONDITION' in disk_file['dataset'][data_name]:
            return condition_type in disk_file['dataset'][data_name]\
                ['CONDITION'].attrs
        else:
            return False

    def set_condition(self, data_name, condition_type, condition_label):
        if self.disk_mode:
            self.__df_set_condition(self.disk_file(), data_name,
                                    condition_type, condition_label)
        else:
            self.__mm_set_condition(data_name, condition_type, condition_label)

    def __mm_set_condition(self, data_name, condition_type, condition_label):
        if not self.has_condition(data_name):
            self.dataset[data_name]['CONDITION'] = {}
        self.__new_condition(condition_type)
        self.dataset[data_name]['CONDITION'][condition_type] = condition_label

    def __df_set_condition(self, disk_file, data_name, condition_type,
                           condition_label):
        if not self.__df_has_condition(disk_file, data_name):
            disk_file['dataset'][data_name].create_group('CONDITION')
        self.__new_condition(condition_type)
        disk_file['dataset'][data_name]['CONDITION'].\
            attrs[condition_type] = condition_label

    def get_condition(self, data_name, condition_type = None):
        if self.disk_mode:
            return self.__df_get_condition(self.disk_file(), data_name,
                                           condition_type)
        else:
            return self.__mm_get_condition(data_name, condition_type)

    def __mm_get_condition(self, data_name, condition_type = None):
        if condition_type is None and self.__mm_has_condition(data_name,
                                                              condition_type):
            return self.dataset[data_name]['CONDITION']
        elif self.has_condition(data_name, condition_type):
            return self.dataset[data_name]['CONDITION'][condition_type]
        else:
            return False

    def __df_get_condition(self, disk_file, data_name, condition_type = None):
        if condition_type is None and \
            self.__df_has_condition(disk_file, data_name, condition_type):
            r = {}
            for condition_type in disk_file['dataset']\
                [data_name]['CONDITION'].attrs:
                r[condition_type] = disk_file['dataset']\
                    [data_name]['CONDITION'].attrs[condition_type]
            return r
        elif self.__df_has_condition(disk_file, data_name, condition_type):
            return disk_file['dataset'][data_name]['CONDITION'].\
                attrs[condition_type]
        else:
            return False

    def del_condition(self, data_name, condition_type = None):
        if self.disk_mode:
            self.__df_del_condition(self.disk_file(), data_name, condition_type)
        else:
            self.__mm_del_condition(data_name, condition_type)

    def __mm_del_condition(self, data_name, condition_type = None):
        if condition_type is None and self.has_condition(data_name):
            self.dataset[data_name].pop('CONDITION')
        elif condition_type is not None and self.has_condition(data_name, condition_type):
            self.dataset[data_name]['CONDITION'].pop(condition_type)

    def __df_del_condition(self, disk_file, data_name, condition_type = None):
        if condition_type is None and self.__df_has_condition(data_name):
            disk_file['dataset'][data_name].pop('CONDITION')
        elif condition_type is not None and \
            self.__df_has_condition(data_name, condition_type):
            disk_file['dataset'][data_name]['CONDITION'].\
                attrs.pop(condition_type)

    ### EPOCH ###

    def __get_epoch(self, data_name, epoch = None):
        if self.disk_mode:
            return self.__df_get_epoch(self.disk_file(), data_name, epoch)
        else:
            return self.__mm_get_epoch(data_name, epoch)

    def __mm_get_epoch(self, data_name, epoch = None):
        if epoch is None:
            return self.__mm_get_epochs(data_name)
        else:
            return self.dataset[data_name]['data'][epoch]

    def __df_get_epoch(self, disk_file, data_name, epoch = None):
        if epoch is None:
            return self.__df_get_epochs(disk_file, data_name)
        else:
            # this returns a `h5py.Group` object
            return disk_file['dataset'][data_name]['data'][str(epoch)]

    def get_epochs(self, data_name):
        if self.disk_mode:
            return self.__df_get_epochs(self.disk_file(), data_name)
        else:
            return self.__mm_get_epochs(data_name)

    def __mm_get_epochs(self, data_name):
        return sorted(list(self.dataset[data_name]['data'].keys()))

    def __df_get_epochs(self, disk_file, data_name):
        return sorted(int(e) for e in 
                      list(disk_file['dataset'][data_name]['data'].keys()))

    def __init_epoch(self, data_name, epoch):
        if self.disk_mode:
            self.__df_init_epoch(self.disk_file(), data_name, epoch)
        else:
            self.__mm_init_epoch(data_name, epoch)

    def __mm_init_epoch(self, data_name, epoch):
        self.dataset[data_name]['data'][epoch] = dict()

    def __df_init_epoch(self, disk_file, data_name, epoch):
        disk_file['dataset'][data_name]['data'].create_group(str(epoch))

    def __del_epoch(self, data_name, epoch):
        if self.disk_mode:
            self.__df_del_epoch(self.disk_file(), data_name, epoch)
        else:
            self.__mm_del_epoch(data_name, epoch)

    def __mm_del_epoch(self, data_name, epoch):
        self.dataset[data_name]['data'].pop(epoch)

    def __df_del_epoch(self, disk_file, data_name, epoch):
        disk_file['dataset'][data_name]['data'].pop(str(epoch))

    def select_epochs(self, element_name=None, element_value=None,
                      data_name=None):
        # data_name
        if data_name is None:
            data_name = self.get_data()
        elif isinstance(data_name, str):
            data_name = [data_name]
        # label_name & label_value
        if element_name is None:
            rst = list()
            for d in data_name:
                if self.__get_state(d) < Dataset.PREPROCESSED:
                    continue
                logging.info(d)
                for epoch in self.get_epochs(d):
                    rst.append((d, epoch))
        elif element_name is not None and element_value is not None:
            rst = list()
            if not isinstance(element_value, list):
                element_value = [element_value]
            for d in data_name:
                if self.__get_state(d) == Dataset.ERROR:
                    continue
                logging.info(d)
                if self.elements[element_name] == 'label':
                    for epoch in self.get_epochs(d):
                        if self.get_label(d, epoch, element_name) in \
                            element_value:
                            rst.append((d, epoch))
                elif self.elements[element_name] == 'condition':
                    if self.get_condition(d, element_name) in element_value:
                        # if matches the condition, add all the epochs into rst
                        for epoch in self.get_epochs(d):
                            rst.append((d, epoch))
        else:
            raise TypeError('You should input all of or none of\n' + 
                            '`element_name` and `element_value`')
        return rst

    def exclude_epochs(self, label_name = None, label_value = None):
        # 删除所有文件中指定label的数据
        if (label_name is not None) and (label_value is not None):
            if not isinstance(label_value, list):
                label_value = [label_value]
            for l in label_value:
                for data_name in self.get_data():
                    if self.__get_state(data_name) == Dataset.ERROR:
                        continue
                    logging.info(data_name)
                    to_be_del = []
                    for epoch in self.get_epochs(data_name):
                        if self.get_label(data_name, epoch, label_name) == l:
                            to_be_del.append(epoch)
                    for e in to_be_del:
                        self.__del_epoch(data_name, e)
        else:
            # When no inputs, all the labels NOT IN LABELDICT will be exclude.
            for label_name in self.labels:
                allowed_label = self.labeldict[label_name].labels()
                for data_name in self.get_data():
                    if self.__get_state(data_name) == Dataset.ERROR:
                        continue
                    logging.info(data_name)
                    to_be_del = []
                    for epoch in self.get_epochs(data_name):
                        if self.get_label(data_name, epoch, label_name) \
                            not in allowed_label:
                            to_be_del.append(epoch)
                    for e in to_be_del:
                        self.__del_epoch(data_name, e)

    def __get_epoch_element_list(self, data_name, epoch):
        if self.disk_mode:
            return self.__df_get_epoch_element_list(self.disk_file(),
                                                    data_name, epoch)
        else:
            return self.__mm_get_epoch_element_list(data_name, epoch)
        
    def __mm_get_epoch_element_list(self, data_name, epoch):
        return list(self.dataset[data_name]['data'][epoch].keys())    

    def __df_get_epoch_element_list(self, disk_file, data_name, epoch):
        r = list(disk_file['dataset'][data_name]['data'][str(epoch)].keys())
        r.extend(list(disk_file['dataset'][data_name]\
                      ['data'][str(epoch)].attrs))
        return r

    ### FEATURE & LABEL ###

    def __set_feature(self, data_name, epoch, feature_name, value):
        '''
        A private method for setting features of dataset. The features are 
        NumPy matrices which represent the data. Most of datasets have \'single 
        feature matrix and single label\'. That means the ML models have single
        input and single output. However, `sleepology` also supports datasets 
        with multiple features and labels. Use different `feature_name` to 
        distinguish different feature matrices.
        
        NOTE: The feature matrices with the SAME feature_name in different 
        epochs or data MUST HAVE THE SAME SHAPE.

        Parameters
        ----------
        data_name : str
            The `data_name` of the data which should be in the 
            `self.dataset.keys()`.
        epoch : int
            The serial number of the epoch in PSG data.
        feature_name : str
            The name for distinguishing different feature matrices. It is 
            necessary in dataset with multiple features. The default is 
            \'FEATURE\'.
        value : np.ndarray or array-like
            The feature matrix.

        Returns
        -------
        None.

        '''
        if self.disk_mode:
            self.__df_set_feature(self.disk_file(), data_name,
                                  epoch, feature_name, value)
        else:
            self.__mm_set_feature(data_name, epoch, feature_name, value)

    def __mm_set_feature(self, data_name, epoch, feature_name, value):
        # initialize a new epoch when the feature value first being set.
        if epoch not in self.get_epochs(data_name):
            self.__init_epoch(data_name, epoch)
        
        self.dataset[data_name]['data'][epoch][feature_name] = value

    def __df_set_feature(self, disk_file, data_name,
                         epoch, feature_name, value):
        if epoch not in self.__df_get_epochs(disk_file, data_name):
            self.__df_init_epoch(disk_file, data_name, epoch)
        if feature_name in disk_file['dataset'][data_name]['data'][str(epoch)]:
            disk_file['dataset'][data_name]['data']\
                [str(epoch)][feature_name][()] = value
        else:
            disk_file['dataset'][data_name]['data'][str(epoch)].\
                create_dataset(feature_name, data = value)

    def get_feature(self, data_name, epoch, feature_name):
        '''
        A private matrix for getting feature matrix of particular data, epoch 
        and feature_name.

        Parameters
        ----------
        data_name : str
            The `data_name` of the data which should be in the 
            `self.dataset.keys()`.
        epoch : int
            The serial number of the epoch in PSG data.
        feature_name : str
            The name for distinguishing different feature matrices. It is 
            necessary in dataset with multiple features. The default is 
            \'FEATURE\'.

        Returns
        -------
        np.ndarray or array-like
            The feature matrix.

        '''
        if self.disk_mode:
            return self.__df_get_feature(self.disk_file(), data_name, 
                                         epoch, feature_name)
        else:
            return self.__mm_get_feature(data_name, epoch, feature_name)

    def __mm_get_feature(self, data_name, epoch, feature_name):
        return np.squeeze(self.dataset[data_name]['data'][epoch][feature_name])


    def __df_get_feature(self, disk_file, data_name, epoch, feature_name):
        rst =  disk_file['dataset'][data_name]['data']\
            [str(epoch)][feature_name][()]
        return np.squeeze(rst)

    def __set_label(self, data_name, epoch, label_name, value):
        '''
        A private method for setting labels of dataset. The labels are 
        NumPy matrices which represent the data. Most of datasets have \'single 
        feature matrix and single label\'. That means the ML models have single
        input and single output. However, `sleepology` also supports datasets 
        with multiple features and labels. Use different `label_name` to 
        distinguish different label matrices.
        
        NOTE: The label matrices with the SAME label_name in different 
        epochs or data MUST HAVE THE SAME SHAPE.

        Parameters
        ----------
        data_name : str
            The `data_name` of the data which should be in the 
            `self.dataset.keys()`.
        epoch : int
            The serial number of the epoch in PSG data.
        label_name : str
            The name for distinguishing different label matrices. It is 
            necessary in dataset with multiple labels. The default is 
            \'LABEL\'.
        value : np.ndarray or array-like
            The label matrix.

        Returns
        -------
        None.

        '''
        if self.disk_mode:
            self.__df_set_label(self.disk_file(), data_name,
                                epoch, label_name, value)
        else:
            self.__mm_set_label(data_name, epoch, label_name, value)

    def __mm_set_label(self, data_name, epoch, label_name, value):
        # initialize a new epoch when the label value first being set.
        if epoch not in self.get_epochs(data_name):
            self.__init_epoch(data_name, epoch)
        self.dataset[data_name]['data'][epoch][label_name] = value

    def __df_set_label(self, disk_file, data_name, epoch, label_name, value):
        if epoch not in self.__df_get_epochs(disk_file, data_name):
            self.__df_init_epoch(disk_file, data_name, epoch)
        disk_file['dataset'][data_name]['data'][str(epoch)].\
            attrs[label_name] = value

    def get_label(self, data_name, epoch, label_name):
        '''
        A private matrix for getting label matrix of particular data, epoch 
        and label_name.

        Parameters
        ----------
        data_name : str
            The `data_name` of the data which should be in the 
            `self.dataset.keys()`.
        epoch : int
            The serial number of the epoch in PSG data.
        label_name : str
            The name for distinguishing different label matrices. It is 
            necessary in dataset with multiple labels. The default is 
            \'LABEL\'.

        Returns
        -------
        np.ndarray or array-like
            The label matrix.

        '''
        if self.disk_mode:
            return self.__df_get_label(self.disk_file(), data_name,
                                       epoch, label_name)
        else:
            return self.__mm_get_label(data_name, epoch, label_name)

    def __mm_get_label(self, data_name, epoch, label_name):
        return self.dataset[data_name]['data'][epoch][label_name]

    def __df_get_label(self, disk_file, data_name, epoch, label_name):
        return disk_file['dataset'][data_name]['data'][str(epoch)].\
            attrs[label_name]

    def del_feature(self, feature_name):
        if self.disk_file:
            self.__df_del_feature(self.disk_file(), feature_name)
        else:
            self.__mm_del_feature(feature_name)

    def __mm_del_feature(self, feature_name):
        if len(self.features) > 1:
            for data_name in self.get_data():
                for epoch in self.get_epochs(data_name):
                    self.dataset[data_name]['data'][epoch].pop(feature_name)
            self.features.remove(feature_name)
        else:
            raise Exception('Cannot delete the only feature.')

    def __df_del_feature(self, disk_file, feature_name):
        if len(self.features) > 1:
            for data_name in self.__df_get_data(disk_file):
                for epoch in self.__df_get_epochs(disk_file, data_name):
                    disk_file['dataset'][data_name]['data'][epoch].\
                        pop(feature_name)
            self.features.remove(feature_name)
        else:
            raise Exception('Cannot delete the only feature.')

    def del_label(self, label_name):
        if self.disk_mode:
            self.__df_del_label(self.disk_file(), label_name)
        else:
            self.__mm_del_label(label_name)

    def __mm_del_label(self, label_name):
        for data_name in self.get_data():
            for epoch in self.get_epochs(data_name):
                self.dataset[data_name]['data'][epoch].pop(label_name)
        self.labels.remove(label_name)
        if len(self.labels) == 0:
            self.has_label = False

    def __df_del_label(self, disk_file, label_name):
        for data_name in self.__df_get_data(disk_file):
            for epoch in self.__df_get_epochs(disk_file, data_name):
                disk_file['dataset'][data_name]['data'][epoch].attrs.\
                    pop(label_name)
        self.labels.remove(label_name)
        if len(self.labels) == 0:
            self.has_label = False

    ### STATISTIC ###

    def stat_condition(self, condition_type, data_name = None):
        rst = {}
        if data_name is None:
            data_name = self.get_data()
        for dn in data_name:
            if self.__get_state(dn) != Dataset.ERROR:
                logging.info(dn)
                l = self.get_condition(dn, condition_type)
                if l:
                    if l not in rst.keys():
                        rst[l] = 0
                    rst[l] += 1
        return rst

    def stat_label(self, label_name, label = None ,data_name = None):
        rst = {}
        if data_name is None:
            data_name = self.get_data()
        if label is None:
            label = self.labeldict[label_name].labels()
        elif not isinstance(label, list):
            label = [label]
        for l in label:
            rst[l] = 0
        for dn in data_name:
            if self.__get_state(dn) != Dataset.ERROR:
                logging.info(dn)
                for epoch in self.get_epochs(dn):
                    l = self.get_label(dn, epoch, label_name)
                    if l in label:
                        rst[l] += 1
        return rst

    def stat_plot_bar(stat, show = True, save_path = None):
        # the input is a dict, whose keys are the `label_name` and
        # values are the count
        x = list(range(len(stat)))
        tick_label = [k for k in stat.keys()]
        y = [stat[k] for k in stat.keys()]
        plt.bar(x,y, align = 'center', tick_label = tick_label)
        plt.xlabel('LABEL')
        plt.ylabel('COUNT')
        if show:
            plt.show()
        if save_path is not None:
            plt.savefig(save_path)

    def stat_plot_pie(stat, show = True, save_path = None):
        # the input is a dict, whose keys are the `label_name` and
        # values are the count
        kinds = [k for k in stat.keys()]
        values = [stat[k] for k in stat.keys()]
        plt.axes(aspect='equal') # ensure the pie is a circle
        plt.pie(values, autopct = '%3.1f%%', labels = kinds)
        if show:
            plt.show()
        if save_path is not None:
            plt.savefig(save_path)

    ### SOURCE ###

    def __df_init_source(self):
        self.sources = {}

    def __df_init_source_data(self, data_name):
        if not hasattr(self, 'sources'):
            self.__df_init_source()
        self.sources[data_name] = {}

    def __df_has_source_data(self, data_name):
        if not hasattr(self, 'sources'):
            return False
        else:
            return data_name in self.sources

    def __set_source(self, data_name, source_name, value):
        if self.disk_mode:
            self.__df_set_source(data_name, source_name, value)
        else:
            self.__mm_set_source(data_name, source_name, value)

    def __mm_set_source(self, data_name, source_name, value):
        self.dataset[data_name]['source'][source_name] = value

    def __df_set_source(self, data_name, source_name, value):
        if not self.__df_has_source_data(data_name):
            self.__df_init_source_data(data_name)
        self.sources[data_name][source_name] = value

    def __has_source(self, data_name, source_name):
        if self.disk_mode:
            return self.__df_has_source(data_name, source_name)
        else:
            return self.__mm_has_source(data_name, source_name)

    def __mm_has_source(self, data_name, source_name):
        rst = source_name in self.dataset[data_name]['source']
        return rst

    def __df_has_source(self, data_name, source_name):
        rst = source_name in self.sources[data_name]
        return rst

    def get_source(self, data_name, source_name = None):
        if self.disk_mode:
            return self.__df_get_source(data_name, source_name)
        else:
            return self.__mm_get_source(data_name, source_name)

    def __mm_get_source(self, data_name, source_name = None):
        if source_name is None:
            return self.dataset[data_name]['source']
        else:
            return self.dataset[data_name]['source'][source_name]

    def __df_get_source(self, data_name, source_name = None):
        if source_name is None:
            return self.sources[data_name]
        else:
            return self.sources[data_name][source_name]
    
    def get_relative_path(self, absolute_path):
        relative_path = os.path.relpath(absolute_path, self.path)
        return relative_path

    def get_absolute_path(self, relative_path):
        absolute_path = os.path.join(self.path, relative_path)
        absolute_path = os.path.normpath(absolute_path)
        return absolute_path

    ### STATE ###

    def __set_state(self, data_name, state = None):
        '''
        A private method for setting the state of data.

        Parameters
        ----------
        data_name : str
            The `data_name` of the data which should be in the 
            `self.dataset.keys()`.
        state : int, optional
            The state to be set. Typically `Dataset.ADDED`, `Dataset.CHECKED`, 
            `Dataset.PREPROCESSED`, `Dataset.PREDICTED` and `Dataset.ERROR`.
            The default is None which will change the state to the next state.
            e.g., changing `Dataset.ADDED` to `Dataset.CHECKED`.

        Returns
        -------
        None.

        '''
        if self.disk_mode:
            self.__df_set_state(self.disk_file(), data_name, state)
        else:
            self.__mm_set_state(data_name, state)
        
    def __mm_set_state(self, data_name, state = None):
        if state is None:
            self.dataset[data_name]['STATE'] += 1
        else:
            self.dataset[data_name]['STATE'] = state

    def __df_set_state(self, disk_file, data_name, state = None):
        if state is None:
            disk_file['dataset'][data_name].attrs['STATE'] += 1
        else:
            disk_file['dataset'][data_name].attrs['STATE'] = state

    def __get_state(self, data_name, return_str = False):
        '''
        A private method for getting the state of data.

        Parameters
        ----------
        data_name : str
            The `data_name` of the data which should be in the 
            `self.dataset.keys()`.
        return_str : bool, optional
            If True, the state will return in type `str`. The default is False 
            which returns `int`.

        Returns
        -------
        int or str
            The state of the data. The type of returns depends on the 
            `return_str` parameter.

        '''
        if self.disk_mode:
            return self.__df_get_state(self.disk_file(), data_name, return_str)
        else:
            return self.__mm_get_state(data_name, return_str)

    def __mm_get_state(self, data_name, return_str = False):
        state = {0 : 'ADDED',
                 1 : 'CHECKED',
                 2 : 'PREPROCESSED',
                 3 : 'PREDICTED',
                 -1: 'ERROR'
                 }
        if return_str:
            return state[self.dataset[data_name].get('STATE')]
        else:
            return self.dataset[data_name].get('STATE')

    def __df_get_state(self, disk_file, data_name, return_str = False):
        state = {0 : 'ADDED',
                 1 : 'CHECKED',
                 2 : 'PREPROCESSED',
                 3 : 'PREDICTED',
                 -1: 'ERROR'
                 }
        if return_str:
            return state[disk_file['dataset'][data_name].attrs['STATE']]
        else:
            return disk_file['dataset'][data_name].attrs['STATE']

    ### PREPROCESS ###

    def prepare_feature(self, data_name, procedure, feature_name = 'FEATURE', 
                        max_epoch = math.inf):
        # 2020-3-1 预处理数据
        method = procedure.procedure()
        i = 0
        while i < math.inf:
            try:
                epoch, value = self.get_source(data_name, feature_name).get()
                value = method(value)
                self.__set_feature(data_name, epoch, feature_name, value)
            except StopIteration:
                break
            except MemoryError:
                # handle the OOM problem
                print('======================================================'
                      '======')
                print('Now we encounter an OUT OF MEMORY error.')
                print('The memory usage of this dataset is approximately'
                      ' {0:.3f} GB'.format(self.memory_usage()))
                print('Now we are at Iteration {0} of \'{1}\'.'\
                      .format(i, data_name))
                print('You can read the log file for more details.')
                print('The possible solution:')
                print('\t- Use a more powerful computer which has larger '
                      'memory.')
                print('\t- Lower your data\'s sampling rate.')
                print('\t- Use less channels.')
                print('\t- Add less data sources.')
                print('======================================================='
                      '=====')
                print('INTERRUPTED')
            except:
                tb = traceback.format_exc()
                self.exception_logger.submit(data_name, i, feature_name, tb)
            finally:
                i += 1
        self.get_source(data_name, feature_name).clean()

    def prepare_label(self, data_name, label_name = 'LABEL', 
                      max_epoch = math.inf):
        # 2020-3-1 预处理标签
        i = 0
        while i < math.inf:
            try:
                epoch, value = self.get_source(data_name, label_name).get()
                self.__set_label(data_name, epoch, label_name, value)
            except StopIteration:
                break
            except:
                tb = traceback.format_exc()
                self.exception_logger.submit(data_name, i, label_name, tb)
            finally:
                i += 1
        self.get_source(data_name, label_name).clean()

    def source_check(self, data_name):
        rst = True
        if self.__get_state(data_name) == Dataset.ERROR:
            rst = False
            return rst
        for f in self.features:
            if not self.__has_source(data_name, f):
                tb = 'Feature ' + f + 'not exist in data ' + data_name + '.'
                self.exception_logger.submit(data_name, None, f, tb)
                rst = False
        if self.has_label:
            for l in self.labels:
                if not self.__has_source(data_name, l):
                    tb = 'Label ' + l + 'not exist in data ' + data_name + '.'
                    self.exception_logger.submit(data_name, None, l, tb)
                    rst = False
        if rst:
            self.__set_state(data_name, Dataset.CHECKED)
        else:
            self.__set_state(data_name, Dataset.ERROR)
        return rst

    def epoch_check(self, data_name):
        rst = True
        to_be_del = []
        if self.__get_state(data_name) == Dataset.ERROR:
            rst = False
            return rst
        elif len(self.get_epochs(data_name)) == 0:
            self.__set_state(data_name, Dataset.ERROR)
            rst = False
            return rst
        for epoch in self.get_epochs(data_name):
            p = False  # pop this epoch or not
            for f in self.features:
                if f not in self.__get_epoch_element_list(data_name, epoch):
                    self.exception_logger.submit(data_name, epoch, f, 
                                                 'Feature is not in dataset.')
                    p = True
            if self.has_label:
                for l in self.labels:
                    if l not in self.__get_epoch_element_list(data_name,
                                                              epoch):
                        self.exception_logger.\
                            submit(data_name, epoch, l, 
                                   'Label is not in dataset.')
                        p = True
            if p:
                to_be_del.append(epoch)
        for e in to_be_del:
            self.__del_epoch(data_name, e)
        if rst:
            self.__set_state(data_name, Dataset.PREPROCESSED)
        else:
            self.__set_state(data_name, Dataset.ERROR)
        return rst

    def shape_check(self):
        # check if every data's state is ERROR
        all_error = True
        for data_name in self.get_data():
            if self.__get_state(data_name) != Dataset.ERROR:
                all_error = False
        if all_error:
            raise Exception('Preprocessing error.')
        # check if features' shape is consistent.
        self.shape = {}
        for data_name in self.get_data():
            # 检查有没有错误，如果没有错误，再进一步检查
            if self.__get_state(data_name) != Dataset.ERROR:
                # 检查shape一致性
                e = False  # data is error or not
                for epoch in self.get_epochs(data_name):
                    # features
                    for f in self.features:
                        if f not in self.shape.keys():
                            self.shape[f] = self.get_feature(data_name, 
                                                          epoch, f).shape
                        else:
                            tb = 'Shape is inconsistent.'
                            if self.shape[f] != self.get_feature(data_name, 
                                                          epoch, f).shape:
                                self.exception_logger.submit(data_name, epoch,
                                                             f, tb)
                                e = True
                if e:
                    self.__set_state(data_name, Dataset.ERROR)

    def preprocess(self, procedure):
        for data_name in self.get_data():
            if self.source_check(data_name):
                if not isinstance(procedure, list):
                    for f in self.features:
                        self.prepare_feature(data_name, procedure, f)
                else:
                    for f, p in procedure:
                        self.prepare_feature(data_name, p, f)
                if self.has_label:
                    for l in self.labels:
                        self.prepare_label(data_name, l)
                self.epoch_check(data_name)
        self.shape_check()

    ### SAMPLE ###

    def stat_classes(self, elements, unit=None):
        '''
        Add up different classes (or class group) according to input elements.
        Return a dict whose `keys()` are classes and `values()` are epochs for
        each class.

        >>> # compute the number of classes
        >>> len(ds.stat_classes(elements).keys())

        >>> # list all the classes
        >>> list(ds.stat_classes(elements).keys())

        Parameters
        ----------
        elements : str, list or dict
            `element_name` or a list of them. Also can be a dict with format:
            >>> {element_name : [element_value, element_value]}
            to add up specific classes.

            For example, you want to add up classes whose `'LABEL'` is '1' or
            '2', and `'DIAGNOSE'` is `'healthy'`. You can use:
            >>> dataset.stat_classes(
            ...    {'LABEL' : ['1', '2'],
            ...     'DIAGNOSE' : 'healthy'})
            
            If you want to add up all the `'LABEL'` but just when `'DIAGNOSE'`
            is `'healthy'`. You can use:
            >>> dataset.stat_classes(
            ...    {'LABEL' : [],  # an empty list means use all the values
            ...     'DIAGNOSE' : 'healthy'})
            Notice that an empty list means use all the values.

            Elements must be `label_name` or `condition_type`.

        Raises
        ------
        TypeError
            Raised when input unsupported parameters.
        ValueError
            Raised when the input element do not exist in the dataset.
            Or Raised when the input element is not `label_name` or
            `condition_type`

        Returns
        -------
        rst : dict
            The key is `str` or `tuple`, represents a class or class group.
            The value is a list of `data_name` or a list of 
            `(data_name, epoch)` tuples.

        '''
        # check input
        select = False  # to control whether to use selected elements or all
        unit_candidate = ['epoch', 'data']
        if unit is not None and unit not in unit_candidate:
            raise ValueError('Invalid unit. Only \'epoch\' or \'data\' allowd')
        if isinstance(elements, str):
            elements = [elements]
        elif isinstance(elements, list):
            pass
        elif isinstance(elements, dict):
            select = True
            for key in elements:
                if isinstance(elements[key], str):
                    elements[key] = [elements[key]]
        else:
            raise TypeError('The input parameter `elements` must be `str`, '
                            '`list` or `dict`. It represents label names or '
                            'condition types.')

        num = len(elements)  # numbers of elements

        # assert elements must be `label` or `condition`
        # and separate labels and conditions
        condition = []
        label = []
        for element in elements:
            if element not in self.elements:
                raise ValueError(element + ' not in this dataset.')
            elif self.elements[element] == 'label':
                label.append(element)
            elif self.elements[element] == 'condition':
                condition.append(element)
            else:
                raise ValueError('The input parameter `elements` must be label'
                                 ' names or condition types.')
        # check passed
        rst = dict()
        if unit is None:
            if len(label) == 0:
                unit = 'data'
            else:
                unit = 'epoch'

        if unit == 'data':
            for d in self.get_data():  # traversal and add up classes
                fit = True
                c = []
                for cond in condition:
                    if select and elements[cond] and self.get_condition(d, cond) \
                        not in elements[cond]:  # encounter an invalid condition
                        fit = False
                        break  # drop this data
                    c.append(self.get_condition(d, cond))
                if not fit:
                    continue
                c = tuple(c) if num > 1 else c[0]  # use tuple or single
                                                   # element
                if c not in rst:
                    rst[c] = []
                rst[c].append(d)
        else:  # the unit is `epoch`
            for d in self.get_data():
                for e in self.get_epochs(d):
                    fit = True
                    c = []
                    for cond in condition:
                        if select and elements[cond] and self.get_condition(d, \
                            cond) not in elements[cond]:  # encounter an invalid condition
                            fit = False
                            break  # drop this epoch
                        c.append(self.get_condition(d, cond))
                    for l in label:
                        if select and elements[l] and self.get_label(d, e, l) \
                            not in elements[l]:  # encounter an invalid label
                            fit = False
                            break  # drop this data
                        c.append(self.get_label(d, e, l))
                    if not fit:
                        continue
                    c = tuple(c) if num > 1 else c[0]  # use tuple or single 
                                                       # element
                    if c not in rst:
                        rst[c] = []
                    rst[c].append((d, e))
        return rst

    def sample_data(self, data_name, lst, tmin=0, tmax=0, data_padding=True,
                    max_len=None, epoch_padding=False, test_data_name=None,
                    autoencoder=False, array_type=int, concat:bool=False,
                    x_dict: Dict[str, BaseDict] = {},
                    y_dict: Dict[str, BaseDict] = {}):
        if test_data_name is None:
            test_data_name = data_name

        x, y = lst
        x_samp = []
        if isinstance(x, str):
            x = [x]
        for i in x:
            x_samp.append(self.sample_serial_x(data_name, i, tmin, tmax,
                                               data_padding, max_len,
                                               epoch_padding,
                                               array_type=array_type,
                                               concat=concat, x_dict=x_dict))
        if len(x_samp) == 1:
            x_samp = x_samp[0]
        else:
            x_samp = np.asarray(x_samp)
        # y
        if autoencoder:
            return (x_samp, x_samp)
        elif y is None:
            return x_samp
        else:
            y_samp = []
            if isinstance(y, str):
                y = [y]
            for i in y:
                y_samp.append(
                    self.__samp_condition(test_data_name, i,
                                         array_type, y_dict))
            if len(y_samp) == 1:
                y_samp = y_samp[0]
            else:
                y_samp = np.asarray(y_samp)
            return (x_samp, y_samp)

    def sample_epoch(self, data_name, epoch, lst, tmin=0, tmax=0,
                    epoch_padding=False, test_data_name=None, test_epoch=None,
                    autoencoder=False, array_type=int, concat:bool=False,
                    x_dict: Dict[str, BaseDict] = {},
                    y_dict: Dict[str, BaseDict] = {}):
        if test_data_name is None:
            test_data_name = data_name
            test_epoch = epoch

        x, y = lst
        x_samp = []
        if isinstance(x, str):
            x = [x]
        for i in x:
            x_samp.append(self.sample_epoched_x(data_name, epoch, i,
                                                tmin, tmax, epoch_padding,
                                                array_type=array_type,
                                                concat=concat, x_dict=x_dict))
        if len(x_samp) == 1:
            x_samp = x_samp[0]
        else:
            x_samp = np.asarray(x_samp)
        # y
        if autoencoder:
            return (x_samp, x_samp)
        elif y is None:
            return x_samp
        else:
            y_samp = []
            if isinstance(y, str):
                y = [y]
            for i in y:
                y_samp.append(self.sample_epoched_y(test_data_name, test_epoch,
                                                    i, array_type=array_type,
                                                    y_dict=y_dict))
            if len(y_samp) == 1:
                y_samp = y_samp[0]
            else:
                y_samp = np.asarray(y_samp)
            return (x_samp, y_samp)

    def sample_epoched_x(self, data_name:str, epoch:int, element_name:str,
                         tmin:int=0, tmax:int=0, padding:bool=False,
                         array_type:type=int, concat:bool=False,
                         x_dict: Dict[str, BaseDict] = {}):
        '''
        A low-level method to sample one epoch according to parameters.

        Parameters
        ----------
        data_name : str
            The target datum.
        epoch : int
            The target epoch.
        element_name : str
            The target element, it may be a feature_name, a label_name or a
            condition_name.
        tmin : int, optional
            The number of preceding epochs before the target epoch to be
            sampled. Usually used for RNN-related model or other models related
            to time series. The default is 0, means do not sample any other
            epochs.
        tmax : int, optional
            The number of following epochs after the target epoch to be
            sampled. Usually used for RNN-related model or other models related
            to time series. The default is 0, means do not sample any other
            epochs.
        padding : bool, optional
            At the beginning or end of the datum, whether the target epoch to
            be padded with zero or dropped. If True, they will be padded, 
            otherwise dropped. The default is False.
        array_type : type or None, optional
            The type of elements in return array. The default and most
            recommended is int. You should input directly a `type`, not a
            `str`.
            For example:
                array_type=bool          (Correct)
                array_type='bool'        (Wrong)

        concat : bool, optional
            Decide the output shape when tmin!=0 or tmax!=0, i.e., multiple
            time steps are sampled.
            If you have not set the `tmin` or `tmax`. You do not need to care
            about this parameter.
            The default is False, means the output will not concatenate time
            steps (i.e. sampled epochs) but leave the output shape as
            `(time_step, channel[, feature-related])`. In this situation, every
            time step occupies a unique matrix. It is suitable for RNN-related
            models.
            If `concat==True`, all the sampled epoches will concatenate by the
            'time' axis. It results in continuous feature which is suitable for
            TCN or other CNN-related models.

        Raises
        ------
        DataStateError
            Raised when your data have not preprocessed yet. Check your data
            state.
        BrokenTimestepError
            The flag of the end of one datum. It is for the iterators. You need
            not do any treatment about it.
        ValueError
            The shape mismatch.

        Returns
        -------
        np.ndarray
            The sampled x.

        '''
        TIMEAXIS = 1  # set the time axis
        # Skip this epoch if returns `None`
        logging.info('== EPOCHED SAMPLE ==')
        # check state
        if self.__get_state(data_name) < Dataset.PREPROCESSED:
            raise DataStateError('You cannot sample from a data not correctly '
                                 'preprocessed. The target data `{0}` has a '
                                 'state `{1}`.'.format(data_name,
                                                       self.__get_state(
                                                           data_name, True)))
        logging.debug('Data_name: {0}'.format(data_name))
        logging.debug('Epoch: {0}'.format(str(epoch)))
        # ===== without timestep =====
        if tmin == 0 and tmax == 0:
            # == feature ==
            if self.elements[element_name] == 'feature':
                return self.get_feature(data_name, epoch, element_name)
            # == label ==
            elif self.elements[element_name] == 'label':
                return self.__samp_label(data_name, epoch,
                                         element_name, array_type, x_dict)
        # ===== with timestep =====
        else:
            # check timespan
            # timespan cannot be longer than data_length
            data_length = len(self.get_epochs(data_name))
            logging.debug('Data Length: ' + str(data_length))
            assert tmax >= 0
            timespan = abs(tmin) + tmax + 1
            assert timespan <= data_length
            # sample
            epoch_list = self.get_epochs(data_name)
            logging.debug('The epoch between [{0}, {1}]'.format(epoch_list[0],
                                                            epoch_list[-1]))
            idx = epoch_list.index(epoch)
            # if `idx - abs(tmin) < 0`, it will cause problem in slicing
            # so check if `idx - abs(tmin) < 0`
            lower = idx - abs(tmin) if idx - abs(tmin) >= 0 else 0
            upper = idx + tmax + 1
            # == feature ==
            if self.elements[element_name] == 'feature':
                dtype = 'float32'
                sample_shape = self.shape[element_name]
                r = np.asarray([self.get_feature(data_name, i, element_name) 
                              for i in epoch_list[lower : upper]])
            # == label ==
            elif self.elements[element_name] == 'label':
                dtype = 'int32'
                sample_shape = self.labeldict[element_name].shape()
                r = np.asarray([self.__samp_label(data_name, i,
                                                  element_name, array_type,
                                                  x_dict)
                        for i in epoch_list[lower : upper]])
            if len(r) < timespan:
                if not padding:
                    raise BrokenTimestepError()
                else:
                    x = np.full((timespan, ) + sample_shape, 0, dtype=dtype)
                    trunc = r[-timespan:]
                    # check `trunc` has expected shape
                    trunc = np.asarray(trunc, dtype=dtype)
                    if trunc.shape[1:] != sample_shape:
                        raise ValueError('Shape of sample {0} of sequence is '
                                         'different from expected shape {1}'.\
                                        format(trunc.shape[1:], sample_shape))
                    if idx - abs(tmin) >= 0:
                        x[:len(trunc)] = trunc
                    elif idx - abs(tmin) < 0:
                        x[-len(trunc):] = trunc
                    r = x
            # == concat ==
            # Only when `x` is a feature, the concat is needed.
            if concat and self.elements[element_name] == 'feature':
                r = np.concatenate(r, axis=TIMEAXIS)
            logging.debug(r.shape)
            return r

    def sample_epoched_y(self, data_name, epoch, element_name, array_type=int,
                         y_dict: Dict[str, BaseDict] = {}):
        # check state
        if self.__get_state(data_name) < Dataset.PREPROCESSED:
            raise DataStateError('You cannot sample from a data not correctly '
                                 'preprocessed. The target data `' + data_name
                                 + '` has a state `' 
                                 + self.__get_state(data_name, True) + '`.')
        # label
        if self.elements[element_name] == 'label':
            return self.__samp_label(data_name, epoch, element_name,
                                     array_type, y_dict)
        elif self.elements[element_name] == 'condition':
            return self.__samp_condition(data_name, element_name,
                                         array_type, y_dict)
        else:
            raise ValueError('The input element ' + element_name + 'is a `'
                             + self.elements[element_name] + '`. It cannot'
                             'act as `epoched_y` sample.')
    
    def sample_serial_x(self, data_name, element_name, tmin, tmax,
                        data_padding, max_len, epoch_padding, array_type=int,
                        concat:bool=False, x_dict: Dict[str, BaseDict] = {}):
        timespan = abs(tmin) + tmax + 1
        rst = []
        for epoch in self.get_epochs(data_name):
            try:
                rst.append(self.sample_epoched_x(data_name, epoch, 
                                                 element_name,
                                                 tmin, tmax, epoch_padding,
                                                 array_type=array_type,
                                                 concat=concat,
                                                 x_dict=x_dict))
            except BrokenTimestepError:
                continue
        if data_padding:
            if max_len is None:
                raise ValueError('A certain `max_len` is needed for '
                                 'data_padding')
            if self.elements[element_name] == 'feature':
                if timespan == 1:
                    sample_shape = self.shape[element_name]
                else:
                    sample_shape = (timespan, ) + self.shape[element_name]
                dtype = 'float32'
            else:
                if timespan == 1:
                    sample_shape = self.labeldict[element_name].shape()
                else:
                    sample_shape = (timespan, ) + \
                        self.labeldict[element_name].shape()
                dtype = 'int32'
            x = np.full((max_len, ) + sample_shape, -1, dtype=dtype)
            trunc = rst[-max_len:]
            # check `trunc` has expected shape
            trunc = np.asarray(trunc, dtype=dtype)
            if trunc.shape[1:] != sample_shape:
                raise ValueError('Shape of sample %s of sequence '
                                 'is different from expected shape %s' %
                                 (trunc.shape[1:], sample_shape))
            x[-len(trunc):] = trunc
            rst = x
        else:
            rst = np.asarray(rst)
        return rst

    def __samp_condition(self, data_name: str, condition_type: str,
                        array_type: type,
                        labeldict: Dict[str, BaseDict] = {}) -> np.ndarray:
        if condition_type in labeldict:
            return labeldict[condition_type].trans(
                self.get_condition(data_name, condition_type),
                array_type=array_type)
        else:
            return self.labeldict[condition_type].trans(
                self.get_condition(data_name, condition_type),
                array_type=array_type) 

    def __samp_label(self, data_name: str, epoch: int, label_type: str,
                     array_type: type,
                     labeldict: Dict[str, BaseDict] = {}) -> Union[int, np.ndarray]:
        if label_type in labeldict:
            return labeldict[label_type].trans(
                self.get_label(data_name, epoch, label_type),
                array_type=array_type)
        else:
            return self.labeldict[label_type].trans(
                self.get_label(data_name, epoch, label_type),
                array_type=array_type)

    def memory_usage(self, unit = 'GB'):
        if self.disk_mode:
            return 0
        else:
            r = total_size(self.dataset)
            if unit.upper() == 'KB':
                r /= 1024
            elif unit.upper() == 'MB':
                r /= 1024 * 1024
            else:
                r /= 1024 * 1024 * 1024
            return r

    ### LOAD & SAVE ###

    def __df_save_attr_0_2(self, disk_file):
        attrs = ['VERSION', 'name', 'comment']
        for attr in attrs:
            if hasattr(self, attr):
                disk_file.attrs[attr] = getattr(self, attr)

    def __df_save_label_dict_0_2(self, disk_file):
        # 保存标签字典
        if self.has_label:
            disk_file.create_group('label_dict')
            for label_name in self.labeldict.keys():
                disk_file['label_dict'].create_group(label_name)
                tem = self.labeldict[label_name].dict
                for label in tem:
                    disk_file['label_dict'][label_name].create_dataset(label, data = tem[label]) # 2020-3-5 存的是int

    def __df_save_dataset_0_2(self, disk_file):
        self.__df_init_dataset(disk_file)
        for data_name in self.get_data():
            if self.__get_state(data_name) == Dataset.PREPROCESSED:
                self.__df_init_new_data(disk_file, data_name)
                self.__df_save_condition_0_2(disk_file, data_name)
                # 保存数据 2020-3-5
                self.__df_save_data_0_2(disk_file, data_name)

    def __df_save_condition_0_2(self, disk_file, data_name):
        l = self.get_condition(data_name)
        if l:
            for condition_type in l.keys():
                self.__df_set_condition(disk_file, data_name, condition_type, l[condition_type])
 
    def __df_save_source_0_2(self, disk_file, data_name):
        # 暂时先不用存source
        disk_file['dataset'][data_name].create_group('source')
        source = self.get_source(data_name)
        for source_name in source.keys():
            disk_file['dataset'][data_name]['source'].create_group(source_name)
            disk_file['dataset'][data_name]['source'].attrs['abs'] = self.get_source(data_name, source_name).path
            disk_file['dataset'][data_name]['source'].attrs['rel'] = self.get_source(data_name, source_name).path
            disk_file['dataset'][data_name]['source'].attrs['source_type'] = self.get_source(data_name, source_name).source_type

    def __df_save_data_0_2(self, disk_file, data_name):
        for epoch in self.get_epochs(data_name):
            self.__df_init_epoch(disk_file, data_name, epoch)
            for feature in self.features:
                self.__df_set_feature(disk_file, data_name, epoch, feature, 
                                      self.get_feature(data_name, epoch, feature))
            # 保存标签 2020-3-5
            if self.has_label:
                for label in self.labels:
                    self.__df_set_label(disk_file, data_name, epoch, label, 
                               self.get_label(data_name, epoch, label))

    def df_load_label_dict_0_2(self, disk_file):
        # 获取标签字典
        if 'label_dict' in disk_file.keys():
            for label_name in disk_file['label_dict'].keys():
                self.labeldict[label_name] = ClassDict()
                for label in disk_file['label_dict'][label_name].keys():
                    self.labeldict[label_name].add(label, int(disk_file['label_dict'][label_name][label][()])) # 2020-3-5 存的是int

    def df_load_dataset_0_2(self, disk_file):
        # 获取数据
        for data_name in disk_file['dataset'].keys():
            self.__init_new_data(data_name)
            self.__df_load_condition_0_2(disk_file, data_name)
            self.__set_state(data_name, Dataset.PREPROCESSED)
            for epoch in self.__df_get_epochs(disk_file, data_name):
                self.__init_epoch(data_name, int(epoch))
                for feature_name in disk_file['dataset'][data_name]['data'][str(epoch)].keys():
                    self.__new_feature(feature_name)
                    self.__set_feature(data_name, int(epoch), feature_name,
                                       self.__df_get_feature(disk_file, data_name, 
                                                      epoch, feature_name))
                for label_name in disk_file['dataset'][data_name]['data'][str(epoch)].attrs:
                    self.has_label = True
                    self.__new_label(label_name)
                    self.__set_label(data_name, int(epoch), label_name,
                                     self.__df_get_label(disk_file, data_name, epoch, 
                                                  label_name))

    def __df_load_condition_0_2(self, disk_file, data_name):
        if self.__df_has_condition(disk_file, data_name):
            l = self.__df_get_condition(disk_file, data_name)
            for condition_type in l.keys():
                self.__new_condition(condition_type)
                self.set_condition(data_name, condition_type, l[condition_type])

    def df_load_source_0_2(self, disk_file, data_name):
        pass

    @staticmethod
    def load(path, mode = 'memory', load_source = False):
        '''
        Load the existed dataset from disk.

        Parameters
        ----------
        path : path-like
            The path to the dataset file.
        mode : {'memory', 'disk'}, optional
            For the faster speed, 'memory' mode will process all the data in
            the memory. But if you have a huge dataset which exceeds or will
            exhaust your memory, the 'disk' mode will cache all the data at
            disk until you are to use them. It do slow down the computation,
            but it is a good resolution when you are not able to extend your
            memory. The default is 'memory'.
        load_source : bool, optional
            Whether to load the sources of dataset. The default is False.

        Returns
        -------
        dataset : sleepology.dataset.Dataset

        '''
        # 2020-3-1 hdf5->dict
        disk_file = h5py.File(path, 'r')
        # 获取属性
        VERSION = disk_file.attrs['VERSION']
        name = disk_file.attrs['name']
        comment = disk_file.attrs['comment']
        dir_name = os.path.dirname(path)
        dataset = Dataset(name, dir_name, comment = comment, mode = mode)
        dataset.VERSION = VERSION
        
        dataset.df_load_label_dict_0_2(disk_file)
        dataset.df_load_dataset_0_2(disk_file)
        
        disk_file.close()
        return dataset

    def save(self, path = None):
        # 2020-2-23 dict->hdf5
        save_path = path if path is not None else self.save_direction
        print('Saved at:')
        print(save_path)
        disk_file = h5py.File(save_path, 'w')
        self.__df_save_attr_0_2(disk_file)
        self.__df_save_label_dict_0_2(disk_file)
        self.__df_save_dataset_0_2(disk_file)
        disk_file.close()


# -*- coding: utf-8 -*-
"""
Created on Tue Jun 30 13:22:57 2020

@author: 赵匡是
"""
import os
import re
import mne
import math
import numpy
import pylsl
import traceback
from xml.etree import ElementTree

mne.set_log_level('WARNING')


class Source(object):
    '''
    要有获取数据的方法
    '''
    def __init__(self, path, source_type):
        self.path = path
        self.source_type = source_type

    def get(self):
        '''
        An abstract method to get next raw data epoch.

        Raises
        ------
        NotImplementedError

        Returns
        -------
        None.

        '''
        raise NotImplementedError('Please implement the method.')


class BCISource(Source):
    def __init__(self, path, source_type, 
                 epoch_length, sample_freq):
        super().__init__(path, source_type)
        self.epoch_length = epoch_length
        self.sample_freq = sample_freq

    


class RawDataFile(BCISource):
    def __init__(self, path, source_type, 
                 epoch_length, sample_freq, name = None):
        super().__init__(path, source_type, 
                         epoch_length, sample_freq)
        if name is None:
            self.name = os.path.basename(self.path)
        else:
            self.name = name

    def get(self, number = None):
        '''
        Get the data of one epoch. The length (time span) of each epoch equals 
        to `epoch_length` you have set when initialize this object.
        You can specify which epoch you want to get by setting parameter 
        `number`.
        When `number is None`, this method will sequentially return epochs 
        until meet the end of data. Then function will raise a `StopIteration`
        exception.
        
        NOTE: After initializing the `RawDataFile` object, when first time 
        using `get` method, it will load data, resample data and split data.
        So the first usage will spend a long time.

        Parameters
        ----------
        number : int, optional
            The serial number of the epoch. You can specify which epoch you 
            want to get. The default is None, means sequentially return epochs 
            until meet the end of data.

        Raises
        ------
        StopIteration
            Encounter the end of data. You may use `try ... except` structure 
            to handle.

        Returns
        -------
        number : int
            The serial number of the epoch.
        epoch : mne.epochs.Epochs

        '''
        if not hasattr(self, 'raw'):
            self.__load_data()
            self.number = 0
        try:
            if number is None:
                number = self.number
                self.number += 1
            return number, self.raw[number]
        except IndexError:
            raise StopIteration
        except AttributeError:
            if not hasattr(self, 'number'):
                raise StopIteration


    def clean(self):
        '''
        Delete the data from memory.

        '''
        del self.raw

    def __load_data(self, args = ()):
        '''
        A private method for load data, resample data and split data into 
        epochs.

        Parameters
        ----------
        args : tuple, optional
            The additional parameters for reading data. The default is ().

        '''
        self.raw = self.__read_raw_data(self.path, self.source_type, args)
        self.raw.load_data()
        self.raw.resample(self.sample_freq)
        events = numpy.array([[i, 0, 1] for i in range(0, self.raw.n_times, int(self.epoch_length * self.raw.info['sfreq']))])
        self.raw = mne.Epochs(self.raw, events, tmin = 0, tmax = self.epoch_length, baseline = None)

    def __read_raw_data(self, fpath, typ, args = ()):
        '''
        A private method for openning raw BCI data files.

        Parameters
        ----------
        fpath : str
            The file path of the raw data file.
        typ : str
            The type of a raw data file. The available type is shown below. 
            You can use the complete type name or the extension in the bracket.
            - BrainVision (vhdr)
            - European data format (edf)
            - BioSemi data format (bdf)
            - General data format (gdf)
            - Neuroscan CNT data format (cnt)
            - EGI simple binary (egi)
            - EGI MFF (mff)
            - EEGLAB set files (set)
            - Nicolet (data)
            - eXimia EEG data (nxe)
        args : tuple, optional
            Additional arguments. Which is depend on file type. You can get 
            detailed usage on 
            https://mne.tools/stable/python_reference.html#reading-raw-data . 
            The default is ().

        Returns
        -------
        mne.io.Raw
        '''
        # 2020-2-21 导入原始数据
        if typ == 'vhdr' or typ == 'BrainVision':
            return mne.io.read_raw_brainvision(fpath, *args)
        elif typ == 'edf' or typ == 'European data format':
            return mne.io.read_raw_edf(fpath, *args) 
        elif typ == 'bdf' or typ == 'BioSemi data format':
            return mne.io.read_raw_bdf(fpath, *args) 
        elif typ == 'gdf' or typ == 'General data format':
            return mne.io.read_raw_gdf(fpath, *args) 
        elif typ == 'cnt' or typ == 'Neuroscan CNT data format':
            return mne.io.read_raw_cnt(fpath, *args) 
        elif typ == 'egi' or typ == 'EGI simple binary':
            return mne.io.read_raw_egi(fpath, *args) 
        elif typ == 'mff' or typ == 'EGI MFF':
            return mne.io.read_raw_egi(fpath, *args) 
        elif typ == 'set' or typ == 'EEGLAB set files':
            return mne.io.read_raw_eeglab(fpath, *args) 
        elif typ == 'data' or typ == 'Nicolet':
            return mne.io.read_raw_nicolet(fpath, *args) 
        elif typ == 'nxe' or typ == 'eXimia EEG data':
            return mne.io.read_raw_eximia(fpath, *args)        



class TCPStream(BCISource):
    # 增加一个self.cursor参数，在get数据时，同步修改cursor。这样可以确定数据的顺序。
    pass


class CSVSource(Source):
    def __init__(self, path, col, sep = ',\s+', source_type = None):
        super().__init__(path, source_type)
        self.col = col
        self.sep = sep

    def get(self, number = None):
        '''
        Get the label of one epoch. 
        You can specify which epoch you want to get by setting parameter 
        `number`.
        When `number is None`, this method will sequentially return epochs 
        until meet the end of data. Then function will raise a `StopIteration`
        exception.

        Parameters
        ----------
        number : int, optional
            The serial number of the epoch. You can specify which epoch you 
            want to get. The default is None, means sequentially return epochs 
            until meet the end of data.

        Raises
        ------
        StopIteration
            Encounter the end of data. You may use `try ... except` structure 
            to handle.

        Returns
        -------
        number : int
            The serial number of the epoch.
        epoch : mne.epochs.Epochs

        '''
        if not hasattr(self, 'raw'):
            self.__load_data()
            self.number = 0
        try:
            if number is None:
                number = self.number
                self.number += 1
            rst = self.raw[number]
            if self.source_type == 'aasm':
                rst = self.trans_aasm(rst)
            return number, rst
        except IndexError:
            raise StopIteration

    def trans_aasm(self, label):
    # 2020-3-2 把标签中的4期睡眠全部转化为3期
        if label == '4':
            label = '3'
        return label

    def clean(self):
        '''
        Delete the data from memory.

        '''
        del self.raw
    
    def __load_data(self):
        with open(self.path, 'r') as f:
            label_raw = f.readlines()     
        self.raw = []
        for entry in label_raw:
            entry = entry.strip()
            self.raw.append(re.split(self.sep, entry)[self.col]) 

class NihonkohdenLabelSource(CSVSource):
    def __init__(self, path, source_type = None):
        super().__init__(path, -3, '\\s+', source_type)

class NsrrLabelSource(Source):
    d = {'Wake|0' : 'W',
        'Stage 1 sleep|1' : '1',
        'Stage 2 sleep|2' : '2',
        'Stage 3 sleep|3' : '3',
        'Stage 4 sleep|4' : '4',
        'REM sleep|5' : 'R',
        'Movement|6' : 'M',
        'Unscored|9' : 'U',
        'Artifact|10' : 'A',
        'Beginning of time in bed' : 'B',
        'End of time in bed' : 'E',
        'Beginning of analysis period' : 'BA',
        'End of analysis period' : 'EA',
        'SpO2 desaturation|SpO2 desaturation' : 'spo2desat',
        'SpO2 artifact|SpO2 artifact' : 'spo2artifact',
        'Hypopnea|Hypopnea' : 'hypopnea',
        'Obstructive hypopnea|Obstructive Hypopnea' : 'obsthypopnea',
        'Central hypopnea|Central Hypopnea' : 'centhypopnea',
        'Mixed hypopnea|Mixed Hypopnea' : 'mixedhypopnea',
        'Obstructive apnea|Obstructive Apnea' : 'obstapnea',
        'Central apnea|Central Apnea' : 'centapnea',
        'Mixed apnea|Mixed Apnea' : 'mixedapnea',
        'Apnea|APNEA' :'apnea'
    }
    def __init__(self, path, label_type='Stages|Stages', source_type=None,
                 epoch_length=30):
        super().__init__(path, source_type)
        self.label_type = label_type
        self.epoch_length = epoch_length

    def get(self, number = None):
        '''
        Get the label of one epoch. 
        You can specify which epoch you want to get by setting parameter 
        `number`.
        When `number is None`, this method will sequentially return epochs 
        until meet the end of data. Then function will raise a `StopIteration`
        exception.

        Parameters
        ----------
        number : int, optional
            The serial number of the epoch. You can specify which epoch you 
            want to get. The default is None, means sequentially return epochs 
            until meet the end of data.

        Raises
        ------
        StopIteration
            Encounter the end of data. You may use `try ... except` structure 
            to handle.

        Returns
        -------
        number : int
            The serial number of the epoch.
        epoch : mne.epochs.Epochs

        '''
        if not hasattr(self, 'raw'):
            self.__load_data()
            self.number = 0
        try:
            if number is None:
                number = self.number
                self.number += 1
            rst = self.raw[number]
            if self.source_type == 'aasm':
                rst = self.trans_aasm(rst)
            return number, rst
        except IndexError:
            raise StopIteration

    def trans_aasm(self, label):
    # 2020-3-2 把标签中的4期睡眠全部转化为3期
        if label == '4':
            label = '3'
        return label

    def clean(self):
        '''
        Delete the data from memory.

        '''
        del self.raw
    
    def __load_data(self):
        label_raw = ElementTree.parse(self.path)
        events = label_raw.findall('ScoredEvents/ScoredEvent')
        self.raw = []
        for entry in events:
            if entry.find('EventType').text == self.label_type:
                self.raw.extend(
                    [NsrrLabelSource.d[entry.find('EventConcept').text]] *
                    int(float(entry.find('Duration').text) / self.epoch_length)
                    )



  


# 2019年3月27日使用，对睡眠脑多导进行分期。
# 网络结构按照cldnn0902实验二的网络结构

import tensorflow as tf;
import json;
import os;
import dataset;
import h5py;

class History(object):
    def __init__(self, save_path, name):
        if not os.path.exists(save_path):
            os.mkdir(save_path)
        self.save_path = save_path
        self.name = name
        self.history = {}
    
    def add(self, fold, history):
        self.history[fold] = history
    
    def save(self):
        pos = os.path.join(self.save_path, self.name + '.history')
        disk_file = h5py.File(pos, 'w')
        for fold in self.history.keys():
            disk_file.create_group(fold)
            for attr in self.history[fold].keys():
                disk_file[fold].create_dataset(attr, 
                         data = self.history[fold][attr])

class Train(object):
    def __init__(self, save_path, name = 'train'):
        self.history = {};
        if not os.path.exists(save_path):
            os.mkdir(save_path);
        self.save_path = save_path;
        self.name = name;
        
    def set_training_param(self, batch_size, learning_rate, training_epoch,
                     epoch_steps, test_steps):
        ## 设置一个batch的大小，请考虑自己计算机的实际性能和数据集的大小
        self.batch_size = batch_size;
        self.learning_rate = learning_rate;
        self.training_epoch = training_epoch;
        self.epoch_steps = epoch_steps;
        self.test_steps = test_steps; # 规定最后测试多少个test batch

    def set_batch_size(self, batch_size):
        self.batch_size = batch_size;
        
    def set_learning_rate(self, learning_rate):
        self.learning_rate = learning_rate;
        
    def set_training_epoch(self, training_epoch):
        self.training_epoch = training_epoch;
    
    def set_epoch_steps(self, epoch_steps):
        self.epoch_steps = epoch_steps;
    
    def set_test_steps(self, test_steps):
        self.test_steps = test_steps;
        
    def __check_shape(self):
        # 2020-3-17 在训练开始前确认所有要素都齐了
        assert self.model.input_shape[1:] == self.data_input_shape;
        assert self.model.output_shape[1:] == self.data_output_shape;
        
    def feed(self, data, n_splits, data_name = None, tmin = 0, tmax = 0, padding = False):
        self.n_splits = n_splits;
        self.feed = data.k_fold(n_splits, data_name, tmin, tmax, padding);
        self.data_input_shape = data.x_input_shape;
        self.data_output_shape = data.y_output_shape;
    
    def set_model(self, model):
        self.model = model;

    def __cross_validation(self):
        pass

    def save(self):
        pos = os.path.join(self.save_path, self.name + '.history');
        disk_file = h5py.File(pos, 'w');
        for fold in self.history.keys():
            disk_file.create_group(fold);
            for attr in self.history[fold].keys():
                disk_file[fold].create_dataset(attr, 
                         data = self.history[fold][attr]);

    def train(self, cross_validation = True): 
        fold_index = 0;
        
        for train_dataset, test_dataset in self.feed.take(self.n_splits if cross_validation else 1):
        # train_dataset是训练集，test_dataset是测试集。train_dataset[0]和test_dataset[0]是data，四维数组；train[1]和test[1]是label，二维数组。
            
            fold_index += 1;
            
            checkpoint_file = os.path.join(self.save_path, 'cp_' + '{:02d}'.format(fold_index) + '{epoch:03d}_{val_loss:.4f}.h5');
            
            train_dataset = tf.data.Dataset.from_tensors(train_dataset).repeat().shuffle(1024).apply(tf.data.experimental.unbatch()).batch(self.batch_size);
            test_dataset = tf.data.Dataset.from_tensors(test_dataset).repeat().shuffle(1024).apply(tf.data.experimental.unbatch()).batch(self.batch_size);
            
            print('lr = ' + str(self.learning_rate) + ';');
            print('batch_size = ' + str(self.batch_size) + ';');
            
            ckpt = tf.keras.callbacks.ModelCheckpoint(checkpoint_file,verbose=1, save_weights_only=True, period=1);
            
            self.model.compile(optimizer = tf.keras.optimizers.Adam(lr = self.learning_rate), loss= 'categorical_crossentropy', metrics = ['accuracy']);
            
            # model.load_weights(os.path.join(save_path, 'test_weights.h5'));
            hist = self.model.fit(train_dataset, epochs = self.training_epoch, callbacks=[ckpt, tf.keras.callbacks.TensorBoard(log_dir = self.save_path)], steps_per_epoch = self.epoch_steps, validation_data = test_dataset, validation_steps = self.test_steps);
            self.history[str(fold_index)] = dict(hist.history);
            
            # Save weights to a HDF5 file
            self.model.save(self.save_path);
            
class NeuralNetworkTrain(Train):
    def __init__(self, save_path, name = 'train'):
        self.history = {};
        if not os.path.exists(save_path):
            os.mkdir(save_path);
        self.save_path = save_path;
        self.name = name;
        
    def set_training_param(self, batch_size, learning_rate, training_epoch,
                     epoch_steps, test_steps):
        ## 设置一个batch的大小，请考虑自己计算机的实际性能和数据集的大小
        self.batch_size = batch_size;
        self.learning_rate = learning_rate;
        self.training_epoch = training_epoch;
        self.epoch_steps = epoch_steps;
        self.test_steps = test_steps; # 规定最后测试多少个test batch

    def set_batch_size(self, batch_size):
        self.batch_size = batch_size;
        
    def set_learning_rate(self, learning_rate):
        self.learning_rate = learning_rate;
        
    def set_training_epoch(self, training_epoch):
        self.training_epoch = training_epoch;
    
    def set_epoch_steps(self, epoch_steps):
        self.epoch_steps = epoch_steps;
    
    def set_test_steps(self, test_steps):
        self.test_steps = test_steps;
        
    def __check_shape(self):
        # 2020-3-17 在训练开始前确认所有要素都齐了
        assert self.model.input_shape[1:] == self.data_input_shape;
        assert self.model.output_shape[1:] == self.data_output_shape;
        
    def feed(self, data, n_splits, data_name = None, tmin = 0, tmax = 0, padding = False):
        self.n_splits = n_splits;
        self.feed = data.k_fold(n_splits, data_name, tmin, tmax, padding);
        self.data_input_shape = data.x_input_shape;
        self.data_output_shape = data.y_output_shape;
    
    def set_model(self, model):
        self.model = model;
                
    def save(self):
        pos = os.path.join(self.save_path, self.name + '.history');
        disk_file = h5py.File(pos, 'w');
        for fold in self.history.keys():
            disk_file.create_group(fold);
            for attr in self.history[fold].keys():
                disk_file[fold].create_dataset(attr, 
                         data = self.history[fold][attr]);

    def train(self, cross_validation = True): 
        fold_index = 0;
        
        for train_dataset, test_dataset in self.feed.take(self.n_splits if cross_validation else 1):
        # train_dataset是训练集，test_dataset是测试集。train_dataset[0]和test_dataset[0]是data，四维数组；train[1]和test[1]是label，二维数组。
            
            fold_index += 1;
            
            checkpoint_file = os.path.join(self.save_path, 'cp_' + '{:02d}'.format(fold_index) + '{epoch:03d}_{val_loss:.4f}.h5');
            
            train_dataset = tf.data.Dataset.from_tensors(train_dataset).repeat().shuffle(1024).apply(tf.data.experimental.unbatch()).batch(self.batch_size);
            test_dataset = tf.data.Dataset.from_tensors(test_dataset).repeat().shuffle(1024).apply(tf.data.experimental.unbatch()).batch(self.batch_size);
            
            print('lr = ' + str(self.learning_rate) + ';');
            print('batch_size = ' + str(self.batch_size) + ';');
            
            ckpt = tf.keras.callbacks.ModelCheckpoint(checkpoint_file,verbose=1, save_weights_only=True, period=1);
            
            self.model.compile(optimizer = tf.keras.optimizers.Adam(lr = self.learning_rate), loss= 'categorical_crossentropy', metrics = ['accuracy']);
            
            # model.load_weights(os.path.join(save_path, 'test_weights.h5'));
            hist = self.model.fit(train_dataset, epochs = self.training_epoch, callbacks=[ckpt, tf.keras.callbacks.TensorBoard(log_dir = self.save_path)], steps_per_epoch = self.epoch_steps, validation_data = test_dataset, validation_steps = self.test_steps);
            self.history[str(fold_index)] = dict(hist.history);
            
            # Save weights to a HDF5 file
            self.model.save(self.save_path);


# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
    name='sleepology',
    version='0.2.65', # TODO: Do not forget changing version!
    url='https://github.com/Zhao-Kuangshi/sleepology',
    author = 'Zhao-Kuangshi',
    author_email = 'contact@zhaokuangshi.cn',
    maintainer = 'Zhao-Kuangshi',
    maintainer_email = 'contact@zhaokuangshi.cn',
    license = 'GPL-3.0 License',
    description = 'A python package to manage, process and sample huge ' \
                  'polysomnogram (PSG) datasets. And it offers functions for '\
                  'machine learning (ML), real-time sampling and closed-loop'\
                  ' stimulation in the future. ',
    packages = ['sleepology'],
    package_dir = {'sleepology': 'sleepology'},
    #py_modules = ['sleepology.__init__',
    #              'sleepology.dataset',
    #              'sleepology.exception_logger',
    #              'sleepology.exceptions',
    #              'sleepology.label_dict',
    #              'sleepology.preprocess',
    #              'sleepology.sample',
    #              'sleepology.source',
    #              'sleepology.utils'],
    package_data = {'sleepology': ['labeltemplate/*.labeltemplate']},
    python_requires = '>=3.5',
    install_requires = ['h5py>=2.10.0',
                        'numpy>=1.16.4',
                        'matplotlib',
                        'datetime',
                        'mne>=0.18.1',
                        'tableprint>=0.9.1',
                        'scikit-learn>=0.19.2',
                        'pylsl']
)


# -*- coding: utf-8 -*-
"""
Created on Sat Mar 21 21:26:18 2020

@author: 赵匡是
"""
import os
import h5py
import numpy
import tensorflow as tf
from abc import ABC, abstractmethod
from packaging import version
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.decomposition import PCA
from sklearn.svm import SVC
from train import History


class Model(object):
    @abstractmethod
    def train(self):
        pass
    
    def validate(self):
        pass
    
    def feed(self, data, n_splits, data_name = None, tmin = 0, tmax = 0, padding = False):
        self.n_splits = n_splits
        self.feed = data.k_fold(n_splits, data_name, tmin, tmax, padding, array_type = None)
        self.data_input_shape = data.x_input_shape
        self.data_output_shape = data.y_output_shape

class LDAModel(Model):
    def __init__(self, model_name):
        self.model_name = model_name
    
    def train(self, save_path, name, cross_validation = True, reshape = None): 
        history = History(save_path, name)
        
        fold_index = 0
        
        for train_dataset, test_dataset in self.feed.take(self.n_splits if cross_validation else 1):
        # train_dataset是训练集，test_dataset是测试集。train_dataset[0]和test_dataset[0]是data，四维数组；train[1]和test[1]是label，二维数组。
            fold_index += 1
            lda_model = LinearDiscriminantAnalysis()
            
            train_data = numpy.array(train_dataset[0])
            train_label = numpy.array(train_dataset[1])
            test_data = numpy.array(test_dataset[0])
            test_label = numpy.array(test_dataset[1])
            
            lda_model.fit(train_data.reshape((train_data.shape[0], -1)), train_label)
            
            # model.load_weights(os.path.join(save_path, 'test_weights.h5'))
            r = {}
            r['acc'] = lda_model.score(train_data.reshape((train_data.shape[0], -1)), train_label)
            print('acc = ' + str(r['acc']))
            r['val_acc'] = lda_model.score(test_data.reshape((test_data.shape[0], -1)), test_label)
            print('val_acc = ' + str(r['val_acc']))
            history.add(str(fold_index), r)
            
            # Save weights to a HDF5 file
            # self.model.save(self.save_path)
        history.save()

class PCALDAModel(Model):
    def __init__(self, model_name, pca_components=None):
        self.model_name = model_name
        self.pca_components = pca_components
    
    def train(self, save_path, name, cross_validation = True, reshape = None): 
        history = History(save_path, name)
        
        fold_index = 0
        
        for train_dataset, test_dataset in self.feed.take(self.n_splits if cross_validation else 1):
        # train_dataset是训练集，test_dataset是测试集。train_dataset[0]和test_dataset[0]是data，四维数组；train[1]和test[1]是label，二维数组。
            fold_index += 1
            pca_model = PCA(n_components=self.pca_components)
            lda_model = LinearDiscriminantAnalysis()
            
            train_data = numpy.array(train_dataset[0])
            train_label = numpy.array(train_dataset[1])
            test_data = numpy.array(test_dataset[0])
            test_label = numpy.array(test_dataset[1])
            
            pca_model.fit(train_data.reshape((train_data.shape[0], -1)))
            train_pc = pca_model.transform(train_data.reshape((train_data.shape[0], -1)))
            lda_model.fit(train_pc, train_label)
            
            # model.load_weights(os.path.join(save_path, 'test_weights.h5'))
            r = {}
            r['acc'] = lda_model.score(train_pc, train_label)
            print('acc = ' + str(r['acc']))
            test_pc = pca_model.transform(test_data.reshape((test_data.shape[0], -1)))
            r['val_acc'] = lda_model.score(test_pc, test_label)
            print('val_acc = ' + str(r['val_acc']))
            history.add(str(fold_index), r)
            
            # Save weights to a HDF5 file
            # self.model.save(self.save_path)
        history.save()

class SVMModel(Model):
    def __init__(self, model_name):
        self.model_name = model_name
    
    def train(self, save_path, name, cross_validation = True, reshape = None): 
        history = History(save_path, name)
        
        fold_index = 0
        
        for train_dataset, test_dataset in self.feed.take(self.n_splits if cross_validation else 1):
        # train_dataset是训练集，test_dataset是测试集。train_dataset[0]和test_dataset[0]是data，四维数组；train[1]和test[1]是label，二维数组。
            fold_index += 1
            svm_model = SVC()
            
            train_data = numpy.array(train_dataset[0])
            train_label = numpy.array(train_dataset[1])
            test_data = numpy.array(test_dataset[0])
            test_label = numpy.array(test_dataset[1])
            
            svm_model.fit(train_data.reshape((train_data.shape[0], -1)), train_label)
            
            # model.load_weights(os.path.join(save_path, 'test_weights.h5'))
            r = {}
            r['acc'] = svm_model.score(train_data.reshape((train_data.shape[0], -1)), train_label)
            print('acc = ' + str(r['acc']))
            r['val_acc'] = svm_model.score(test_data.reshape((train_data.shape[0], -1)), test_label)
            print('val_acc = ' + str(r['val_acc']))
            history.add(str(fold_index), r)
            
            # Save weights to a HDF5 file
            # self.model.save(self.save_path)
        history.save()

class NeuralNetworkModel(Model, tf.keras.Model):
    def __init__(self, model_name):
        super(NeuralNetworkModel, self).__init__()
        self.model_name = model_name
        
    
    def save(self, filepath, overwrite=True, include_optimizer=True, 
             save_format='h5', signatures=None, options=None):
        filepath = os.path.join(filepath, self.model_name + '.model')
        if version.parse(tf.__version__) < version.parse('2.0.0'):
            super(SequentialNeuralNetworkModel, self).save(filepath, overwrite, 
             include_optimizer)
        else:
            super(SequentialNeuralNetworkModel, self).save(filepath, overwrite, 
             include_optimizer, save_format, signatures, options)
    
    @staticmethod
    def load(filepath, custom_objects=None, compile=True):
        
        model = tf.keras.models.load_model(filepath, 
                                           custom_objects, compile)
        model.__class__ = NeuralNetworkModel
        model_name = os.path.splitext(os.path.basename(filepath))[0]
        model.model_name = model_name
        return model
    
    def set_training_param(self, batch_size, learning_rate, training_epoch,
                     epoch_steps, test_steps):
        ## 设置一个batch的大小，请考虑自己计算机的实际性能和数据集的大小
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.training_epoch = training_epoch
        self.epoch_steps = epoch_steps
        self.test_steps = test_steps # 规定最后测试多少个test batch

    def set_batch_size(self, batch_size):
        self.batch_size = batch_size
        
    def set_learning_rate(self, learning_rate):
        self.learning_rate = learning_rate
        
    def set_training_epoch(self, training_epoch):
        self.training_epoch = training_epoch
    
    def set_epoch_steps(self, epoch_steps):
        self.epoch_steps = epoch_steps
    
    def set_test_steps(self, test_steps):
        self.test_steps = test_steps
    
    def feed(self, data, n_splits, data_name = None, tmin = 0, tmax = 0, padding = False):
        self.n_splits = n_splits
        self.feed = data.k_fold(n_splits, data_name, tmin, tmax, padding)
        self.data_input_shape = data.x_input_shape
        self.data_output_shape = data.y_output_shape
        
    
    def train(self, save_path, name, cross_validation = True): 
        history = History(save_path, name)
        
        fold_index = 0
        if version.parse(tf.__version__) < version.parse('1.13'):
            self.compile(optimizer = tf.train.AdamOptimizer(learning_rate = self.learning_rate), loss= 'categorical_crossentropy', metrics = ['accuracy'])
        else:
            self.compile(optimizer = tf.keras.optimizers.Adam(lr = self.learning_rate), loss= 'categorical_crossentropy', metrics = ['accuracy'])
        
        for train_dataset, test_dataset in self.feed.take(self.n_splits if cross_validation else 1):
        # train_dataset是训练集，test_dataset是测试集。train_dataset[0]和test_dataset[0]是data，四维数组；train[1]和test[1]是label，二维数组。
            
            fold_index += 1
            
            checkpoint_file = os.path.join(save_path, 'cp_' + '{:02d}'.format(fold_index) + '{epoch:03d}_{val_loss:.4f}.h5')
            
            train_dataset = tf.data.Dataset.from_tensors(train_dataset).repeat().shuffle(1024).apply(tf.data.experimental.unbatch()).batch(self.batch_size)
            test_dataset = tf.data.Dataset.from_tensors(test_dataset).repeat().shuffle(1024).apply(tf.data.experimental.unbatch()).batch(self.batch_size)
            
            print('lr = ' + str(self.learning_rate) + '')
            print('batch_size = ' + str(self.batch_size) + '')
            
            ckpt = tf.keras.callbacks.ModelCheckpoint(checkpoint_file,verbose=1, save_weights_only=True, period=1)
            
            # model.load_weights(os.path.join(save_path, 'test_weights.h5'))
            hist = self.fit(train_dataset, epochs = self.training_epoch, callbacks=[ckpt, tf.keras.callbacks.TensorBoard(log_dir = save_path)], steps_per_epoch = self.epoch_steps, validation_data = test_dataset, validation_steps = self.test_steps)
            history.add(str(fold_index), dict(hist.history))
            
            # Save weights to a HDF5 file
            self.save(save_path)
        
        history.save()
        
class SequentialNeuralNetworkModel(tf.keras.Sequential):
    def __init__(self, model_name):
        super(SequentialNeuralNetworkModel, self).__init__()
        self.model_name = model_name
    
    def save(self, filepath, overwrite=True, include_optimizer=True, 
             save_format='h5', signatures=None, options=None):
        filepath = os.path.join(filepath, self.model_name + '.model')
        if version.parse(tf.__version__) < version.parse('2.0.0'):
            super(SequentialNeuralNetworkModel, self).save(filepath, overwrite, 
             include_optimizer)
        else:
            super(SequentialNeuralNetworkModel, self).save(filepath, overwrite, 
             include_optimizer, save_format, signatures, options)
    
    @staticmethod
    def load(filepath, custom_objects=None, compile=True):
        
        model = tf.keras.models.load_model(filepath, 
                                           custom_objects, compile)
        model.__class__ = SequentialNeuralNetworkModel
        model_name = os.path.splitext(os.path.basename(filepath))[0]
        model.model_name = model_name
        return model
    
    def set_training_param(self, batch_size, learning_rate, training_epoch,
                     epoch_steps, test_steps):
        ## 设置一个batch的大小，请考虑自己计算机的实际性能和数据集的大小
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.training_epoch = training_epoch
        self.epoch_steps = epoch_steps
        self.test_steps = test_steps # 规定最后测试多少个test batch

    def set_batch_size(self, batch_size):
        self.batch_size = batch_size
        
    def set_learning_rate(self, learning_rate):
        self.learning_rate = learning_rate
        
    def set_training_epoch(self, training_epoch):
        self.training_epoch = training_epoch
    
    def set_epoch_steps(self, epoch_steps):
        self.epoch_steps = epoch_steps
    
    def set_test_steps(self, test_steps):
        self.test_steps = test_steps
    
    def feed(self, data, n_splits, data_name = None, tmin = 0, tmax = 0, padding = False):
        self.n_splits = n_splits
        self.feed = data.k_fold(n_splits, data_name, tmin, tmax, padding)
        self.data_input_shape = data.x_input_shape
        self.data_output_shape = data.y_output_shape
        
    
    def train(self, save_path, name, cross_validation = True): 
        history = History(save_path, name)
        
        fold_index = 0
        if version.parse(tf.__version__) < version.parse('1.13'):
            self.compile(optimizer = tf.train.AdamOptimizer(learning_rate = self.learning_rate), loss= 'categorical_crossentropy', metrics = ['accuracy'])
        else:
            self.compile(optimizer = tf.keras.optimizers.Adam(lr = self.learning_rate), loss= 'categorical_crossentropy', metrics = ['accuracy'])
        
        for train_dataset, test_dataset in self.feed.take(self.n_splits if cross_validation else 1):
        # train_dataset是训练集，test_dataset是测试集。train_dataset[0]和test_dataset[0]是data，四维数组；train[1]和test[1]是label，二维数组。
            
            fold_index += 1
            
            checkpoint_file = os.path.join(save_path, 'cp_' + '{:02d}'.format(fold_index) + '{epoch:03d}_{val_loss:.4f}.h5')
            
            train_dataset = tf.data.Dataset.from_tensors(train_dataset).repeat().shuffle(1024).apply(tf.data.experimental.unbatch()).batch(self.batch_size)
            test_dataset = tf.data.Dataset.from_tensors(test_dataset).repeat().shuffle(1024).apply(tf.data.experimental.unbatch()).batch(self.batch_size)
            
            print('lr = ' + str(self.learning_rate) + '')
            print('batch_size = ' + str(self.batch_size) + '')
            
            ckpt = tf.keras.callbacks.ModelCheckpoint(checkpoint_file,verbose=1, save_weights_only=True, period=1)
            
            # model.load_weights(os.path.join(save_path, 'test_weights.h5'))
            hist = self.fit(train_dataset, epochs = self.training_epoch, callbacks=[ckpt, tf.keras.callbacks.TensorBoard(log_dir = save_path)], steps_per_epoch = self.epoch_steps, validation_data = test_dataset, validation_steps = self.test_steps)
            history.add(str(fold_index), dict(hist.history))
            
            # Save weights to a HDF5 file
            self.save(save_path)
        
        history.save()
